(* Author: Vincent Cremet *)
(* Date: July 2007 *) 
(* File: SimplyTypedLambdaCalculus_normalization.v *)
(* Description: Simply typed lambda-calculus with pairs *)

(** * Simply typed lambda-calculus *)

Require Import List.
Require Import Arith.
Require Import Omega.

Require Import Util.
Require Import MatchCompNat.
Require Import SimplyTypedLambdaCalculus.

Module LambdaCalculusNormalization(Constants: LambdaConstants).

Module My_LambdaCalculus := LambdaCalculus(Constants).

Import My_LambdaCalculus.
Import Constants.

Fixpoint polySubst_in_term (t: LTerm) (k: nat) (us': (list LTerm)) {struct us'}: LTerm :=
  match us' with
    | nil => t
    | (u :: us) => (polySubst_in_term (subst_in_term t (k + (length us)) u) k us)
  end
.

(** Poly-substitution preserves typing **)
Lemma PolySubstitution_preserves_typing:
  forall (G1 Us G2: LEnv) (T: LType) (t: LTerm) (us: (list LTerm)),
    (Forall2 (fun u U => (WellTyped G1 u U)) us Us) ->
    (WellTyped (G1 ++ Us ++ G2) t T) ->
    (WellTyped (G1 ++ G2) (polySubst_in_term t (length G2) us) T)
.

Proof.
intros. generalize t Us H0 H. clear t Us H0 H. induction us.
(* nil *)
intros. simpl. inversion H. rewrite <- H2 in H0. simpl in H0. trivial.
(* u :: us *)
rename a into u. intros.
simpl. inversion H. clear xA xsA H1 H2.
eapply IHus with (Us := xsB).
rewrite plus_comm.
replace (length us) with (length xsB).
rewrite <- length_of_append.
eapply Substitution_preserves_typing. apply H3.
replace (xB :: xsB ++ G2) with ((xB :: xsB) ++ G2).
rewrite H4. trivial. trivial.
symmetry. eapply Forall2_implies_same_length. apply H5. trivial.
Qed.

(** Poly-substitution preserves typing (corollary 1) **)
Lemma PolySubstitution_preserves_typing_cor1:
  forall (G1 Us: LEnv) (t: LTerm) (T: LType) (us: (list LTerm)),
    (WellTyped (G1 ++ Us) t T) ->
    (Forall2 (fun u U => (WellTyped G1 u U)) us Us) ->
    (WellTyped G1 (polySubst_in_term t 0 us) T)
.

Proof.
intros.
replace 0 with (length (nil (A := LType))).
replace G1 with (G1 ++ (nil (A := LType))).
eapply PolySubstitution_preserves_typing. apply H0.
replace (Us ++ nil) with Us. trivial.
rewrite <- app_nil_end. trivial.
rewrite <- app_nil_end. trivial.
trivial.
Qed.

(***************************************************************)

(** Relation: Reductibility *)
Fixpoint Reductible (G: LEnv) (t: LTerm) (T': LType) {struct T'}: Prop :=
  (WellTyped G t T') /\
  (Acc (symRel RedTerm) t) /\
  match T' with 

    | LType_base =>
      (Acc (symRel RedTerm) t)

    | (LType_fun T U) =>
      forall (G': LEnv) (u: LTerm),
        (Reductible (G' ++ G) u T) ->
        (Reductible (G' ++ G) (LTerm_app t u) U)      

    | (LType_prod T U) =>
      (Reductible G (LTerm_fst t) T) /\
      (Reductible G (LTerm_snd t) U)
  end
.

(*************************************************************)

Lemma Reductible_implies_WellTyped:
  forall (G: LEnv) (t: LTerm) (T: LType),
    (Reductible G t T) ->
    (WellTyped G t T)
.
Proof.
intros. destruct T. simpl in H. tauto. simpl in H. tauto. simpl in H. tauto.
Qed.

Lemma Reductible_implies_SN:
  forall (G: LEnv) (t: LTerm) (T: LType),
    (Reductible G t T) ->
    (Acc (symRel RedTerm) t)
.
Proof.
intros. destruct T. simpl in H. tauto. simpl in H. tauto. simpl in H. tauto.
Qed.

Lemma Reductible2:
  forall (G: LEnv) (t: LTerm),
    (Reductible G t LType_base) ->
    (Acc (symRel RedTerm) t)
.
Proof. intros. simpl in H. tauto. Qed.

Lemma Reductible3:
  forall (G: LEnv) (t: LTerm) (T U: LType),
    (Reductible G t (LType_fun T U)) ->
    forall (G': LEnv) (u: LTerm),
      (Reductible (G' ++ G) u T) ->
      (Reductible (G' ++ G) (LTerm_app t u) U)
.
Proof. intros. simpl in H. destruct H. destruct H1. auto. Qed.

Lemma Reductible3_cor:
  forall (G: LEnv) (t u: LTerm) (T U: LType),
    (Reductible G t (LType_fun T U)) ->
    (Reductible G u T) ->
    (Reductible G (LTerm_app t u) U)
.
Proof.
intros. replace G with (nil ++ G). eapply Reductible3. eauto. trivial. trivial.
Qed.

Lemma Reductible4:
  forall (G: LEnv) (t: LTerm) (T U: LType),
    (Reductible G t (LType_prod T U)) ->
    (Reductible G (LTerm_fst t) T)
.
Proof. intros. simpl in H. tauto. Qed.

Lemma Reductible5:
  forall (G: LEnv) (t: LTerm) (T U: LType),
    (Reductible G t (LType_prod T U)) ->
    (Reductible G (LTerm_snd t) U)
.
Proof. intros. simpl in H. tauto. Qed.

Hint Resolve Reductible_implies_WellTyped Reductible_implies_SN Reductible2
             Reductible3 Reductible3_cor Reductible4 Reductible5 : reductible_db.

Lemma Reductible_base:
  forall (G: LEnv) (t: LTerm),
    (WellTyped G t LType_base) ->
    (Acc (symRel RedTerm) t) ->
    (Reductible G t LType_base)
.

Proof.
intros. simpl. tauto.
Qed.

Lemma Reductible_fun:
  forall (G: LEnv) (t: LTerm) (T U: LType),
    (WellTyped G t (LType_fun T U)) ->
    (Acc (symRel RedTerm) t) ->
    (forall (G': LEnv) (u: LTerm),
      (Reductible (G' ++ G) u T) ->
      (Reductible (G' ++ G) (LTerm_app t u) U)
    ) ->
    (Reductible G t (LType_fun T U))
.

Proof.
intros. simpl. tauto.
Qed.

Lemma Reductible_prod:
  forall (G: LEnv) (t: LTerm) (T U: LType),
    (WellTyped G t (LType_prod T U)) ->
    (Acc (symRel RedTerm) t) ->
    (Reductible G (LTerm_fst t) T) ->
    (Reductible G (LTerm_snd t) U) ->
    (Reductible G t (LType_prod T U))
.

Proof.
intros. simpl. tauto.
Qed.

(*****************************************************************)

Lemma Reductible_left_weakening:
  forall (G G': LEnv) (t: LTerm) (T: LType),
    (Reductible G t T) ->
    (Reductible (G' ++ G) t T)
.

Proof.
intros. generalize G' t H. clear G' t H. induction T.
(* base *)
intros. apply Reductible_base. apply Weakening_of_well_typing_by_the_left. 
auto with reductible_db.
eauto with reductible_db.
(* fun *)
intros. apply Reductible_fun. apply Weakening_of_well_typing_by_the_left.
auto with reductible_db. eauto with reductible_db.
intro G'0.
replace (G'0 ++ G' ++ G) with ((G'0 ++ G') ++ G). eauto with reductible_db.
auto with datatypes.
(* prod *)
intros. apply Reductible_prod. apply Weakening_of_well_typing_by_the_left.
eauto with reductible_db. eauto with reductible_db.
apply IHT1. eauto with reductible_db.
apply IHT2. eauto with reductible_db.
Qed.

(************************************************************************)

Lemma SN_var_intro:
  forall (i: nat),
    (Acc (symRel RedTerm) (LTerm_var i))
.

Proof.
intros.
apply Acc_intro. intros. unfold symRel in H. inversion H.
Qed.

(*********************************************************************)

Lemma polySubst_in_const:
  forall (c: LConst) (k: nat) (us: (list LTerm)),
    (polySubst_in_term (LTerm_const c) k us) =
    (LTerm_const c)
.

Proof.
intros. induction us.
(* nil *)
simpl. trivial.
(* cons *)
simpl.
rewrite IHus. trivial.
Qed.

Lemma polySubst_in_fun:
  forall (T: LType) (t: LTerm) (k: nat) (us: (list LTerm)),
    (polySubst_in_term (LTerm_fun T t) k us) =
    (LTerm_fun T (polySubst_in_term t (S k) us))
.

Proof.
intros. generalize T t k. clear T t k. induction us.
(* nil *)
intros. simpl. trivial.
(* cons *)
intros. simpl.
replace (S (k + length us)) with ((S k) + (length us)).
rewrite IHus. trivial. trivial.
Qed.

Lemma polySubst_in_app:
  forall (t u: LTerm) (k: nat) (us: (list LTerm)),
    (polySubst_in_term (LTerm_app t u) k us) =
    (LTerm_app (polySubst_in_term t k us)
               (polySubst_in_term u k us))
.

Proof.
intros. generalize t u k. clear t u k. induction us.
(* nil *)
intros. simpl. trivial.
(* cons *)
intros. simpl. rewrite IHus. trivial.
Qed.

Lemma polySubst_in_pair:
  forall (t u: LTerm) (k: nat) (us: (list LTerm)),
    (polySubst_in_term (LTerm_pair t u) k us) =
    (LTerm_pair (polySubst_in_term t k us)
                (polySubst_in_term u k us))
.

Proof.
intros. generalize t u k. clear t u k. induction us.
(* nil *)
intros. simpl. trivial.
(* cons *)
intros. simpl. rewrite IHus. trivial.
Qed.

Lemma polySubst_in_fst:
  forall (t: LTerm) (k: nat) (us: (list LTerm)),
    (polySubst_in_term (LTerm_fst t) k us) =
    (LTerm_fst (polySubst_in_term t k us))
.

Proof.
intros. generalize t k. clear t k. induction us.
(* nil *)
intros. simpl. trivial.
(* cons *)
intros. simpl. rewrite IHus. trivial.
Qed.

Lemma polySubst_in_snd:
  forall (t: LTerm) (k: nat) (us: (list LTerm)),
    (polySubst_in_term (LTerm_snd t) k us) =
    (LTerm_snd (polySubst_in_term t k us))
.

Proof.
intros. generalize t k. clear t k. induction us.
(* nil *)
intros. simpl. trivial.
(* cons *)
intros. simpl. rewrite IHus. trivial.
Qed.

Lemma polySubst_in_var1:
  forall (us: (list LTerm)) (i: nat),
    i >= (length us) ->
    (polySubst_in_term (LTerm_var i) 0 us) = (LTerm_var (i - (length us)))
.

Proof.
intros. generalize i H. clear i H. induction us.
(* nil *)
intros. simpl.
replace (i - 0) with i. trivial. omega.
(* cons *)
intros. simpl in H. simpl.
unfold subst_in_var. rewrite match_comp_nat3_gt.
rewrite IHus. replace (pred i - length us) with (i - S (length us)). trivial.
omega. omega. omega.
Qed.

Lemma polySubst_by_snoc:
  forall (t u: LTerm) (k: nat) (us: (list LTerm)),
    (polySubst_in_term t k (snoc us u)) =
    (subst_in_term (polySubst_in_term t (S k) us) k u)
.

Proof.
intros. generalize t u k. clear t u k.
induction us.
(* nil *)
intros. simpl. replace (k + 0) with k. trivial. omega.
(* cons *)
intros. simpl. rewrite length_of_append. simpl.
replace (k + (length us + 1)) with (S (k + length us)).
rewrite <- IHus. trivial. omega.
Qed.

Lemma polySubst_of_var_invariant:
  forall (i k: nat) (us: (list LTerm)), 
    i < k ->
    (polySubst_in_term (LTerm_var i) k us) = (LTerm_var i)
.

Proof.
intros. generalize i k H. clear i k H. induction us.
(* nil *)
intros. simpl. trivial.
(* cons *)
intros. simpl. unfold subst_in_var. rewrite match_comp_nat3_lt.
rewrite IHus. trivial. trivial. omega.
Qed.

Lemma lift_in_var_by_zero:
  forall (i k: nat),
    (lift_at_depth_in_var i k 0) = i
.

Proof.
intros.
destruct (match_comp_nat2_aux_0 i k).
unfold lift_at_depth_in_var. rewrite match_comp_nat2_lt. trivial. trivial.
unfold lift_at_depth_in_var. rewrite match_comp_nat2_ge. omega. trivial.
Qed.

Lemma lift_in_term_by_zero:
  forall (t: LTerm) (k: nat),
    (lift_at_depth_in_term t k 0) = t
.

Proof.
intro t. induction t.
intros. simpl. trivial.
intros. simpl. rewrite lift_in_var_by_zero. trivial.
intros. simpl. rewrite IHt. trivial.
intros. simpl. rewrite IHt1. rewrite IHt2. trivial.
intros. simpl. rewrite IHt1. rewrite IHt2. trivial.
intros. simpl. rewrite IHt. trivial.
intros. simpl. rewrite IHt. trivial.
Qed.

Lemma Permuting_lifting_and_polySubstitution:
  forall (t: LTerm) (i: nat) (us: (list LTerm)) (n: nat),
    (lift_in_term (polySubst_in_term t i us) n) =
    (polySubst_in_term (lift_in_term t n) (i + n) us)
.

Proof.
intros. generalize i. clear i.
pattern us. apply snoc_ind.
(* nil *)
intros. simpl. trivial.
(* snoc *)
intros.
rewrite polySubst_by_snoc.
rewrite Permuting_type_lifting_and_substitution.
rewrite H.
replace (S i + n) with (S (i + n)).
rewrite <- polySubst_by_snoc. trivial. trivial.
Qed.

Lemma polySubst_in_var_variant:
  forall (u: LTerm) (us: (list LTerm)) (i: nat),
    (List_from us i u) ->
    (polySubst_in_term (LTerm_var i) 0 us) = u
.

Proof.
intros. generalize u i H. clear u i H. 
pattern us. apply snoc_ind.
(* nil *)
intros. inversion H.
absurd (snoc xs x = nil). apply snoc_ind_aux_3. trivial.
absurd (snoc xs x = nil). apply snoc_ind_aux_3. trivial.
(* snoc *)
intros. rewrite polySubst_by_snoc. inversion H0.
(* - List-from-head *)
rewrite polySubst_of_var_invariant. simpl. unfold subst_in_var.
rewrite match_comp_nat3_eq. unfold lift_in_term.
rewrite lift_in_term_by_zero.
destruct (snoc_inject _ _ _ _ _ H2).
rewrite <- H5. trivial.
trivial. omega.
(* - List-from-tail *)
replace (LTerm_var (S n)) with (lift_at_depth_in_term (LTerm_var n) 0 1).
replace (polySubst_in_term (lift_at_depth_in_term (LTerm_var n) 0 1) 1 xs) with
  (lift_at_depth_in_term (polySubst_in_term (LTerm_var n) 0 xs) 0 1).
assert (List_from xs n u). replace xs with xs0. trivial.
destruct (snoc_inject _ _ _ _ _ H1); trivial.
rewrite (H _ _ H5).
fold (lift_in_term u 1).
replace 1 with (S (0 + 0)).
rewrite Type_lifting_stronger_than_substitution. simpl.
unfold lift_in_term. apply lift_in_term_by_zero.
trivial.
fold (lift_in_term (LTerm_var n) 1).
fold (lift_in_term (polySubst_in_term (LTerm_var n) 0 xs) 1).
pattern 1 at 3. replace 1 with (0 + 1).
apply  Permuting_lifting_and_polySubstitution. trivial.
simpl. unfold lift_at_depth_in_var. rewrite match_comp_nat2_ge.
replace (n + 1) with (S n). trivial. omega. omega.
Qed.

(*********************************************************************)

Lemma Reductible_subject_reduction:
  forall (T: LType) (G: LEnv) (t t': LTerm),
    (Reductible G t T) ->
    (RedTerm t t') ->
    (Reductible G t' T)
.

Proof.
intro T. induction T.
(* base *)
intros. apply Reductible_base. eapply Subject_reduction.
apply Reductible_implies_WellTyped. apply H. trivial.
simpl in H. destruct H. destruct H1.
inversion H1. apply H3. trivial.
(* fun *)
intros.
assert (WellTyped G t' (LType_fun T1 T2)).
simpl in H. destruct H. eapply Subject_reduction. apply H. trivial.
assert (Acc (symRel RedTerm) t').
simpl in H. destruct H. destruct H2. inversion H2. apply H4. trivial.
apply Reductible_fun. trivial. trivial.
intros. apply IHT2 with (t := (LTerm_app t u)).
simpl in H. destruct H. destruct H4.
apply H5. trivial.
apply R_app_left. trivial.
(* prod *)
intros.
assert (WellTyped G t' (LType_prod T1 T2)). 
simpl in H. destruct H. eapply Subject_reduction. apply H. trivial.
assert (Acc (symRel RedTerm) t').
simpl in H. destruct H. destruct H2. inversion H2. apply H4. trivial.
simpl in H. destruct H. destruct H3. destruct H4.
apply Reductible_prod. trivial. trivial.
apply IHT1 with (t := (LTerm_fst t)). trivial. apply R_fst_cong. trivial.
apply IHT2 with (t := (LTerm_snd t)). trivial. apply R_snd_cong. trivial.
Qed.

Lemma Reductible_subject_reduction_star:
  forall (T: LType) (G: LEnv) (t t': LTerm),
    (Reductible G t T) ->
    (RedTermStar t t') ->
    (Reductible G t' T)
.

Proof.
intros. induction H0.
eapply Reductible_subject_reduction; eauto. trivial. tauto.
Qed.

(************************************************************************)

Module Type PreReductible.

Parameter PreRed: LEnv -> LTerm -> LType -> Prop.

Parameter PreRed_implies_WellTyped:
  forall (G: LEnv) (t: LTerm) (T: LType),
    (PreRed G t T) ->
    (WellTyped G t T)
.

Parameter PreRed_left_weakening:
  forall (G G': LEnv) (t: LTerm) (T: LType),
    (PreRed G t T) ->
    (PreRed (G' ++ G) t T)
.

Parameter PreRed_implies_SN:
  forall (G: LEnv) (t: LTerm) (T: LType),
    (PreRed G t T) ->
    (Acc (symRel RedTerm) t)
.

Parameter PreRed_subject_reduction:
  forall (G: LEnv) (t t': LTerm) (T: LType),
    (PreRed G t T) ->
    (RedTerm t t') ->
    (PreRed G t' T) \/ (Reductible G t' T)
.

Parameter PreRed_not_a_function:
  forall (G: LEnv) (t: LTerm) (T: LType),
    (PreRed G t T) ->
    forall (U: LType) (u: LTerm),
      t <> (LTerm_fun U u)
.

Parameter PreRed_not_a_pair:
  forall (G: LEnv) (t: LTerm) (T: LType),
    (PreRed G t T) ->
    forall (u1 u2: LTerm),
      t <> (LTerm_pair u1 u2)
.

End PreReductible.

Module ProveReductible(PR: PreReductible).

Import PR.

Inductive ClosRed (G: LEnv): LTerm -> LType -> Prop :=

  | ClosRed_base:
    forall (t: LTerm) (T: LType),
      (PreRed G t T) ->
      (ClosRed G t T)

  | ClosRed_app:
    forall (t u: LTerm) (T U: LType),
      (ClosRed G t (LType_fun T U)) ->
      (Reductible G u T) ->
      (ClosRed G (LTerm_app t u) U)

  | ClosRed_fst:
    forall (t: LTerm) (T U: LType),
      (ClosRed G t (LType_prod T U)) ->
      (ClosRed G (LTerm_fst t) T)

  | ClosRed_snd:
    forall (t: LTerm) (T U: LType),
      (ClosRed G t (LType_prod T U)) ->
      (ClosRed G (LTerm_snd t) U)
.

Lemma ClosRed_left_weakening:
  forall (G G': LEnv) (t: LTerm) (T: LType),
    (ClosRed G t T) ->
    (ClosRed (G' ++ G) t T)
.

Proof.
intros. induction H.
apply ClosRed_base. apply PreRed_left_weakening; trivial.
eapply ClosRed_app. eauto. apply Reductible_left_weakening; trivial.
eapply ClosRed_fst. eauto.
eapply ClosRed_snd. eauto.
Qed.

Lemma ClosRed_implies_WellTyped:
  forall (G: LEnv) (t: LTerm) (T: LType),
    (ClosRed G t T) ->
    (WellTyped G t T)
.

Proof.
intros. induction H.
(* base *)
eapply PreRed_implies_WellTyped. trivial.
(* app *)
eapply WT_app. eauto. eauto with reductible_db.
(* fst *)
eapply WT_fst. eauto.
(* snd *)
eapply WT_snd. eauto.
Qed.

Lemma ClosRed_not_a_function:
  forall (G: LEnv) (t: LTerm) (T: LType),
    (ClosRed G t T) ->
    forall (U: LType) (u: LTerm),
      t <> (LTerm_fun U u)
.

Proof.
intros. inversion H.
eapply PreRed_not_a_function; eauto.
intro. discriminate H4.
intro. discriminate H3.
intro. discriminate H3.
Qed.

Lemma ClosRed_not_a_pair:
  forall (G: LEnv) (t: LTerm) (T: LType),
    (ClosRed G t T) ->
    forall (u1 u2: LTerm),
      t <> (LTerm_pair u1 u2)
.

Proof.
intros. inversion H.
eapply PreRed_not_a_pair; eauto.
intro. discriminate H4.
intro. discriminate H3.
intro. discriminate H3.
Qed.

Lemma ClosRed_subject_reduction:
  forall (G: LEnv) (t t': LTerm) (T: LType),
    (ClosRed G t T) ->
    (RedTerm t t') ->
    (ClosRed G t' T) \/ (Reductible G t' T)
.

Proof.
intros. generalize t' H0. clear t' H0. induction H.
(* base *)
intros. destruct (PreRed_subject_reduction _ _ _ _ H H0).
left. apply ClosRed_base; trivial. right. trivial.
(* app *)
intros. inversion H1.
(* R-app-left *)
destruct (IHClosRed _ H5).
left. eapply ClosRed_app. eauto. trivial.
right. eauto with reductible_db.
(* R-app-right *)
left. eapply ClosRed_app. eauto. eapply Reductible_subject_reduction; eauto.
(* R-app-redex *)
absurd (t = LTerm_fun T0 t0). eapply ClosRed_not_a_function; eauto. symmetry; trivial.
(* fst *)
intros. inversion H0.
(* R-fst-cong *)
destruct (IHClosRed _ H2).
left. eapply ClosRed_fst. eauto.
right. eauto with reductible_db.
(* R-fst-redex *)
absurd (t = LTerm_pair t0 u). eapply ClosRed_not_a_pair; eauto. symmetry; trivial.
(* snd *)
intros. inversion H0.
(* R-snd-cong *)
destruct (IHClosRed _ H2).
left. eapply ClosRed_snd. eauto.
right. eauto with reductible_db.
(* R-snd-redex *)
absurd (t = LTerm_pair t0 u). eapply ClosRed_not_a_pair; eauto. symmetry; trivial.
Qed.

Lemma ClosRed_implies_SN_aux1:
  forall (G: LEnv) (t u: LTerm) (T U: LType),
    (ClosRed G t (LType_fun T U)) ->
    (Acc (symRel RedTerm) t) ->
    (Reductible G u T) ->
    (Acc (symRel RedTerm) u) ->
    (Acc (symRel RedTerm) (LTerm_app t u))
.

Proof.
intros. generalize u T U H H1 H2. clear u T U H H1 H2. induction H0.
intros. generalize T U H1 H2. clear T U H1 H2. induction H3.

intros. apply Acc_intro. intros. unfold symRel in H5. inversion H5.
(* R-app-left *)
destruct (ClosRed_subject_reduction _ _ _ _ H3 H9).
eapply H0. trivial. eauto. trivial. eauto with reductible_db.
eauto with reductible_db.
(* R-app-right*)
eapply H2. trivial. eauto. eapply Reductible_subject_reduction; eauto.
(* R-app-redex *)
absurd (x = LTerm_fun T0 t). eapply ClosRed_not_a_function; eauto. symmetry; trivial.
Qed.

Lemma ClosRed_implies_SN_aux2:
  forall (G: LEnv) (t: LTerm) (T U: LType),
    (ClosRed G t (LType_prod T U)) ->
    (Acc (symRel RedTerm) t) ->
    (Acc (symRel RedTerm) (LTerm_fst t))
.

Proof.
intros. generalize T U H. clear T U H. induction H0.

intros. apply Acc_intro. intros. unfold symRel in H2. inversion H2.
(* R-fst-cong *)
destruct (ClosRed_subject_reduction _ _ _ _ H1 H4).
eapply H0. trivial. eauto.
eauto with reductible_db.
(* R-fst-redex *)
absurd (x = LTerm_pair t u). eapply ClosRed_not_a_pair; eauto. symmetry; trivial.
Qed.

Lemma ClosRed_implies_SN_aux3:
  forall (G: LEnv) (t: LTerm) (T U: LType),
    (ClosRed G t (LType_prod T U)) ->
    (Acc (symRel RedTerm) t) ->
    (Acc (symRel RedTerm) (LTerm_snd t))
.

Proof.
intros. generalize T U H. clear T U H. induction H0.

intros. apply Acc_intro. intros. unfold symRel in H2. inversion H2.
(* R-snd-cong *)
destruct (ClosRed_subject_reduction _ _ _ _ H1 H4).
eapply H0. trivial. eauto.
eauto with reductible_db.
(* R-snd-redex *)
absurd (x = LTerm_pair t u). eapply ClosRed_not_a_pair; eauto. symmetry; trivial.
Qed.

Theorem ClosRed_implies_SN:
  forall (G: LEnv) (t: LTerm) (T: LType),
    (ClosRed G t T) ->
    (Acc (symRel RedTerm) t)
.

Proof.
intros. induction H.
(* base *)
eapply PreRed_implies_SN. eauto.
(* app *)
eapply ClosRed_implies_SN_aux1. eauto. trivial. trivial. eauto with reductible_db.
(* fst *)
eapply ClosRed_implies_SN_aux2. eauto. trivial.
(* snd *)
eapply ClosRed_implies_SN_aux3. eauto. trivial.
Qed.

Lemma ClosRed_implies_Reductible:
  forall (G: LEnv) (t: LTerm) (T: LType),
    (ClosRed G t T) ->
    (Reductible G t T)
.

Proof.
intros. generalize G t H. clear G t H. induction T.
(* base *)
intros. simpl. split.
apply ClosRed_implies_WellTyped; trivial.
split. eapply ClosRed_implies_SN; eauto. eapply ClosRed_implies_SN; eauto.
(* fun *)
intros. apply Reductible_fun.
apply ClosRed_implies_WellTyped; trivial.
eapply ClosRed_implies_SN; eauto.
intros.
apply IHT2. eapply ClosRed_app.
apply ClosRed_left_weakening; eauto.
eauto.
(* prod *)
intros. apply Reductible_prod.
apply ClosRed_implies_WellTyped; trivial.
eapply ClosRed_implies_SN; eauto.
apply IHT1. eapply ClosRed_fst. eauto.
apply IHT2. eapply ClosRed_snd. eauto.
Qed.

Lemma PreRed_implies_Reductible:
  forall (G: LEnv) (t: LTerm) (T: LType),
    (PreRed G t T) ->
    (Reductible G t T)
.

Proof.
intros.
apply ClosRed_implies_Reductible. apply ClosRed_base. trivial.
Qed.

End ProveReductible.

(************************************************************************)
(* Constants are reductible *)
(************************************************************************)

Inductive PreRed_const (G: LEnv): LTerm -> LType -> Prop :=

  | PreRed_const_intro:
    forall (c: LConst) (T: LType),
      (WellTyped G (LTerm_const c) T) ->
      (PreRed_const G (LTerm_const c) T)
.

Module PreReductible_const. (* with Variable PreRed := PreRed_const. *)

Definition PreRed: LEnv -> LTerm -> LType -> Prop := PreRed_const.

Lemma PreRed_implies_WellTyped:
  forall (G: LEnv) (t: LTerm) (T: LType),
    (PreRed G t T) ->
    (WellTyped G t T)
.

Proof.
intros. unfold PreRed in H. destruct H. trivial.
Qed.

Lemma PreRed_left_weakening:
  forall (G G': LEnv) (t: LTerm) (T: LType),
    (PreRed G t T) ->
    (PreRed (G' ++ G) t T)
.

Proof.
unfold PreRed. intros. destruct H. apply PreRed_const_intro.
apply Weakening_of_well_typing_by_the_left; trivial.
Qed.

Lemma PreRed_implies_SN:
  forall (G: LEnv) (t: LTerm) (T: LType),
    (PreRed G t T) ->
    (Acc (symRel RedTerm) t)
.

Proof.
intros. unfold PreRed in H. destruct H. apply Acc_intro. intros. inversion H0.
Qed.

Lemma PreRed_subject_reduction:
  forall (G: LEnv) (t t': LTerm) (T: LType),
    (PreRed G t T) ->
    (RedTerm t t') ->
    (PreRed G t' T) \/ (Reductible G t' T)
.

Proof.
intros. unfold PreRed in H. destruct H. inversion H0.
Qed.

Lemma PreRed_not_a_function:
  forall (G: LEnv) (t: LTerm) (T: LType),
    (PreRed G t T) ->
    forall (U: LType) (u: LTerm),
      t <> (LTerm_fun U u)
.

Proof.
intros. unfold PreRed in H. destruct H. intro. discriminate H0.
Qed.

Lemma PreRed_not_a_pair:
  forall (G: LEnv) (t: LTerm) (T: LType),
    (PreRed G t T) ->
    forall (u1 u2: LTerm),
      t <> (LTerm_pair u1 u2)
.

Proof.
intros. unfold PreRed in H. destruct H. intro. discriminate H0.
Qed.

End PreReductible_const.

Module ProveReductible_const := ProveReductible(PreReductible_const).

Lemma Constants_are_reductible:
  forall (G: LEnv) (c: LConst) (T: LType),
    (WellTyped G (LTerm_const c) T) ->
    (Reductible G (LTerm_const c) T)
.

Proof.
intros.
apply ProveReductible_const.PreRed_implies_Reductible.
unfold PreReductible_const.PreRed. apply PreRed_const_intro; trivial.
Qed.

(************************************************************************)
(* Variables are reductible *)
(************************************************************************)

Inductive PreRed_var (G: LEnv): LTerm -> LType -> Prop :=

  | PreRed_var_intro:
    forall (i: nat) (T: LType),
      (WellTyped G (LTerm_var i) T) ->
      (PreRed_var G (LTerm_var i) T)
.

Module PreReductible_var. (* with Variable PreRed := PreRed_var. *)

Definition PreRed: LEnv -> LTerm -> LType -> Prop := PreRed_var.

Lemma PreRed_implies_WellTyped:
  forall (G: LEnv) (t: LTerm) (T: LType),
    (PreRed G t T) ->
    (WellTyped G t T)
.

Proof.
intros. unfold PreRed in H. destruct H. trivial.
Qed.

Lemma PreRed_left_weakening:
  forall (G G': LEnv) (t: LTerm) (T: LType),
    (PreRed G t T) ->
    (PreRed (G' ++ G) t T)
.

Proof.
unfold PreRed. intros. destruct H. apply PreRed_var_intro.
apply Weakening_of_well_typing_by_the_left; trivial.
Qed.

Lemma PreRed_implies_SN:
  forall (G: LEnv) (t: LTerm) (T: LType),
    (PreRed G t T) ->
    (Acc (symRel RedTerm) t)
.

Proof.
intros. unfold PreRed in H. destruct H. apply Acc_intro. intros. inversion H0.
Qed.

Lemma PreRed_subject_reduction:
  forall (G: LEnv) (t t': LTerm) (T: LType),
    (PreRed G t T) ->
    (RedTerm t t') ->
    (PreRed G t' T) \/ (Reductible G t' T)
.

Proof.
intros. unfold PreRed in H. destruct H. inversion H0.
Qed.

Lemma PreRed_not_a_function:
  forall (G: LEnv) (t: LTerm) (T: LType),
    (PreRed G t T) ->
    forall (U: LType) (u: LTerm),
      t <> (LTerm_fun U u)
.

Proof.
intros. unfold PreRed in H. destruct H. intro. discriminate H0.
Qed.

Lemma PreRed_not_a_pair:
  forall (G: LEnv) (t: LTerm) (T: LType),
    (PreRed G t T) ->
    forall (u1 u2: LTerm),
      t <> (LTerm_pair u1 u2)
.

Proof.
intros. unfold PreRed in H. destruct H. intro. discriminate H0.
Qed.

End PreReductible_var.

Module ProveReductible_var := ProveReductible(PreReductible_var).

Lemma Variables_are_reductible:
  forall (G: LEnv) (i: nat) (T: LType),
    (WellTyped G (LTerm_var i) T) ->
    (Reductible G (LTerm_var i) T)
.

Proof.
intros.
apply ProveReductible_var.PreRed_implies_Reductible.
unfold PreReductible_var.PreRed. apply PreRed_var_intro; trivial.
Qed.

(************************************************************************)
(* First projections are reductible *)
(************************************************************************)

Inductive PreRed_fstProj (G: LEnv): LTerm -> LType -> Prop :=

  | PreRed_fstProj_intro:
    forall (t u: LTerm) (T U: LType),
      (Reductible G t T) ->
      (WellTyped G u U) ->
      (Acc (symRel RedTerm) u) ->
      (PreRed_fstProj G (LTerm_fst (LTerm_pair t u)) T)
.

Module PreReductible_fstProj. (* with Variable PreRed := PreRed_fstProj. *)

Definition PreRed: LEnv -> LTerm -> LType -> Prop := PreRed_fstProj.

Lemma PreRed_implies_WellTyped:
  forall (G: LEnv) (t: LTerm) (T: LType),
    (PreRed G t T) ->
    (WellTyped G t T)
.

Proof.
intros. unfold PreRed in H. destruct H.
eapply WT_fst. eapply WT_pair. eauto with reductible_db. eauto.
Qed.

Lemma PreRed_left_weakening:
  forall (G G': LEnv) (t: LTerm) (T: LType),
    (PreRed G t T) ->
    (PreRed (G' ++ G) t T)
.

Proof.
unfold PreRed. intros. destruct H. eapply PreRed_fstProj_intro.
apply Reductible_left_weakening; trivial.
eapply Weakening_of_well_typing_by_the_left; eauto. trivial.
Qed.

Lemma PreRed_implies_SN:
  forall (G: LEnv) (t: LTerm) (T: LType),
    (PreRed G t T) ->
    (Acc (symRel RedTerm) t)
.

Proof.
cut (
  forall (G: LEnv) (t u: LTerm) (T U: LType),
    (Reductible G t T) ->
    (Acc (symRel RedTerm) t) ->
    (WellTyped G u U) ->
    (Acc (symRel RedTerm) u) ->
    (Acc (symRel RedTerm) (LTerm_fst (LTerm_pair t u)))
).
intros. unfold PreRed in H0. destruct H0.
eapply H; eauto with reductible_db.
intros. generalize u T U H H1 H2. clear u T U H H1 H2. induction H0.
intros. generalize T U H1 H2. clear T U H1 H2. induction H3.
intros. apply Acc_intro. intros. unfold symRel in H5. inversion H5. inversion H7.
(* *)
eapply H0. trivial. eapply Reductible_subject_reduction; eauto. eauto.
apply Acc_intro; trivial.
(* *)
eapply H2. trivial. eauto. eapply Subject_reduction; eauto.
(* *)
rewrite <- H6. eauto with reductible_db.
Qed.

Lemma PreRed_subject_reduction:
  forall (G: LEnv) (t t': LTerm) (T: LType),
    (PreRed G t T) ->
    (RedTerm t t') ->
    (PreRed G t' T) \/ (Reductible G t' T)
.

Proof.
intros. unfold PreRed in H. destruct H. inversion H0. inversion H4.
left. eapply PreRed_fstProj_intro.
eapply Reductible_subject_reduction; eauto. eauto. trivial.
left. eapply PreRed_fstProj_intro.
eauto. eapply Subject_reduction; eauto. inversion H2.  apply H10. trivial.
right. rewrite <- H3. trivial.
Qed.

Lemma PreRed_not_a_function:
  forall (G: LEnv) (t: LTerm) (T: LType),
    (PreRed G t T) ->
    forall (U: LType) (u: LTerm),
      t <> (LTerm_fun U u)
.

Proof.
intros. unfold PreRed in H. destruct H. intro. discriminate H2.
Qed.

Lemma PreRed_not_a_pair:
  forall (G: LEnv) (t: LTerm) (T: LType),
    (PreRed G t T) ->
    forall (u1 u2: LTerm),
      t <> (LTerm_pair u1 u2)
.

Proof.
intros. unfold PreRed in H. destruct H. intro. discriminate H2.
Qed.

End PreReductible_fstProj.

Module ProveReductible_fstProj := ProveReductible(PreReductible_fstProj).

Lemma First_projections_are_reductible:
  forall (G: LEnv) (t u: LTerm) (T U: LType),
    (Reductible G t T) ->
    (WellTyped G u U) ->
    (Acc (symRel RedTerm) u) ->
    (Reductible G (LTerm_fst (LTerm_pair t u)) T)
.

Proof.
intros.
apply ProveReductible_fstProj.PreRed_implies_Reductible.
unfold PreReductible_fstProj.PreRed. eapply PreRed_fstProj_intro; eauto.
Qed.

(************************************************************************)
(* Second projections are reductible *)
(************************************************************************)

Inductive PreRed_sndProj (G: LEnv): LTerm -> LType -> Prop :=

  | PreRed_sndProj_intro:
    forall (t u: LTerm) (T U: LType),
      (WellTyped G t T) ->
      (Acc (symRel RedTerm) t) ->
      (Reductible G u U) ->
      (PreRed_sndProj G (LTerm_snd (LTerm_pair t u)) U)
.

Module PreReductible_sndProj. (* with Variable PreRed := PreRed_sndProj. *)

Definition PreRed: LEnv -> LTerm -> LType -> Prop := PreRed_sndProj.

Lemma PreRed_implies_WellTyped:
  forall (G: LEnv) (t: LTerm) (T: LType),
    (PreRed G t T) ->
    (WellTyped G t T)
.

Proof.
intros. unfold PreRed in H. destruct H.
eapply WT_snd. eapply WT_pair. eauto. eauto with reductible_db.
Qed.

Lemma PreRed_left_weakening:
  forall (G G': LEnv) (t: LTerm) (T: LType),
    (PreRed G t T) ->
    (PreRed (G' ++ G) t T)
.

Proof.
unfold PreRed. intros. destruct H. eapply PreRed_sndProj_intro.
eapply Weakening_of_well_typing_by_the_left; eauto. trivial.
apply Reductible_left_weakening; trivial.
Qed.

Lemma PreRed_implies_SN:
  forall (G: LEnv) (t: LTerm) (T: LType),
    (PreRed G t T) ->
    (Acc (symRel RedTerm) t)
.

Proof.
cut (
  forall (G: LEnv) (t u: LTerm) (T U: LType),
    (WellTyped G t T) ->
    (Acc (symRel RedTerm) t) ->
    (Reductible G u U) ->
    (Acc (symRel RedTerm) u) ->
    (Acc (symRel RedTerm) (LTerm_snd (LTerm_pair t u)))
).
intros. unfold PreRed in H0. destruct H0.
eapply H; eauto with reductible_db.

intros. generalize t T U H H0 H1. clear t T U H H0 H1. induction H2.
intros. generalize T U H1 H3. clear T U H1 H3. induction H2.
intros. apply Acc_intro. intros. unfold symRel in H5. inversion H5. inversion H7.
(* *)
eapply H2. trivial. eapply Subject_reduction; eauto. eauto.
(* *)
eapply H0. trivial. eauto. 
apply Acc_intro; trivial.
eapply Reductible_subject_reduction; eauto.
(* *)
rewrite <- H6. eauto with reductible_db.
Qed.

Lemma PreRed_subject_reduction:
  forall (G: LEnv) (t t': LTerm) (T: LType),
    (PreRed G t T) ->
    (RedTerm t t') ->
    (PreRed G t' T) \/ (Reductible G t' T)
.

Proof.
intros. unfold PreRed in H. destruct H. inversion H0. inversion H4.
left. eapply PreRed_sndProj_intro.
eapply Subject_reduction; eauto. inversion H1.  apply H10. trivial. trivial.
left. eapply PreRed_sndProj_intro.
eauto. trivial. eapply Reductible_subject_reduction; eauto.
right. rewrite <- H3. trivial.
Qed.

Lemma PreRed_not_a_function:
  forall (G: LEnv) (t: LTerm) (T: LType),
    (PreRed G t T) ->
    forall (U: LType) (u: LTerm),
      t <> (LTerm_fun U u)
.

Proof.
intros. unfold PreRed in H. destruct H. intro. discriminate H2.
Qed.

Lemma PreRed_not_a_pair:
  forall (G: LEnv) (t: LTerm) (T: LType),
    (PreRed G t T) ->
    forall (u1 u2: LTerm),
      t <> (LTerm_pair u1 u2)
.

Proof.
intros. unfold PreRed in H. destruct H. intro. discriminate H2.
Qed.

End PreReductible_sndProj.

Module ProveReductible_sndProj := ProveReductible(PreReductible_sndProj).

Lemma Second_projections_are_reductible:
  forall (G: LEnv) (t u: LTerm) (T U: LType),
    (WellTyped G t T) ->
    (Acc (symRel RedTerm) t) ->
    (Reductible G u U) ->
    (Reductible G (LTerm_snd (LTerm_pair t u)) U)
.

Proof.
intros.
apply ProveReductible_sndProj.PreRed_implies_Reductible.
unfold PreReductible_sndProj.PreRed. eapply PreRed_sndProj_intro; eauto.
Qed.

(************************************************************************)
(* Lambda redices are reductible *)
(************************************************************************)

Inductive PreRed_appFun (G: LEnv): LTerm -> LType -> Prop :=

  | PreRed_appFun_intro:
    forall (t u: LTerm) (T U: LType),
      (WellTyped (snoc G T) t U) ->
      (Acc (symRel RedTerm) t) ->
      (Reductible G u T) ->
      (Reductible G (subst_in_term t 0 u) U) ->
      (PreRed_appFun G (LTerm_app (LTerm_fun T t) u) U)
.

Module PreReductible_appFun. (* with Variable PreRed := PreRed_appFun. *)

Definition PreRed: LEnv -> LTerm -> LType -> Prop := PreRed_appFun.

Lemma PreRed_implies_WellTyped:
  forall (G: LEnv) (t: LTerm) (T: LType),
    (PreRed G t T) ->
    (WellTyped G t T)
.

Proof.
intros. unfold PreRed in H. destruct H.
eapply WT_app. eapply WT_fun. trivial. eauto with reductible_db.
Qed.

Lemma PreRed_left_weakening:
  forall (G G': LEnv) (t: LTerm) (T: LType),
    (PreRed G t T) ->
    (PreRed (G' ++ G) t T)
.

Proof.
unfold PreRed. intros. destruct H. eapply PreRed_appFun_intro.
rewrite snoc_of_append. eapply Weakening_of_well_typing_by_the_left; eauto. trivial.
apply Reductible_left_weakening; trivial.
apply Reductible_left_weakening; trivial.
Qed.

Lemma PreRed_implies_SN:
  forall (G: LEnv) (t: LTerm) (T: LType),
    (PreRed G t T) ->
    (Acc (symRel RedTerm) t)
.

Proof.
cut (
  forall (G: LEnv) (t u: LTerm) (T U: LType),
    (WellTyped (snoc G T) t U) ->
    (Acc (symRel RedTerm) t) ->
    (Reductible G u T) ->
    (Acc (symRel RedTerm) u) ->
    (Reductible G (subst_in_term t 0 u) U) ->
    (Acc (symRel RedTerm) (LTerm_app (LTerm_fun T t) u))
).
intros. unfold PreRed in H0. destruct H0.
eapply H; eauto with reductible_db.

intros. generalize u T U H H1 H2 H3. clear u T U H H1 H2 H3. induction H0.
intros. generalize T U H1 H2 H4. clear T U H1 H2 H4. induction H3.
intros. apply Acc_intro. intros. unfold symRel in H6. inversion H6. inversion H10.
(* *)
eapply H0. trivial.
eapply Subject_reduction; eauto. trivial.
apply Acc_intro; trivial.
eapply Reductible_subject_reduction. apply H5. apply Substitution_in_reduction. trivial.
(* *)
eapply H2. trivial. eauto. 
eapply Reductible_subject_reduction; eauto.
eapply Reductible_subject_reduction_star. apply H5.
eapply Reduction_in_substitution. trivial.
(* *)
eauto with reductible_db.
Qed.

Lemma PreRed_subject_reduction:
  forall (G: LEnv) (t t': LTerm) (T: LType),
    (PreRed G t T) ->
    (RedTerm t t') ->
    (PreRed G t' T) \/ (Reductible G t' T)
.

Proof.
intros. unfold PreRed in H. destruct H. inversion H0. inversion H7.
(* *)
left. eapply PreRed_appFun_intro.
eapply Subject_reduction; eauto. inversion H1. apply H12; trivial. trivial.
eapply Reductible_subject_reduction. apply H3. apply Substitution_in_reduction. trivial.
(* *)
left. eapply PreRed_appFun_intro.
trivial. trivial. 
eapply Reductible_subject_reduction; eauto.
eapply Reductible_subject_reduction_star. apply H3.
eapply Reduction_in_substitution. trivial.
(* *)
right. trivial.
Qed.

Lemma PreRed_not_a_function:
  forall (G: LEnv) (t: LTerm) (T: LType),
    (PreRed G t T) ->
    forall (U: LType) (u: LTerm),
      t <> (LTerm_fun U u)
.

Proof.
intros. unfold PreRed in H. destruct H. intro. discriminate H3.
Qed.

Lemma PreRed_not_a_pair:
  forall (G: LEnv) (t: LTerm) (T: LType),
    (PreRed G t T) ->
    forall (u1 u2: LTerm),
      t <> (LTerm_pair u1 u2)
.

Proof.
intros. unfold PreRed in H. destruct H. intro. discriminate H3.
Qed.

End PreReductible_appFun.

Module ProveReductible_appFun := ProveReductible(PreReductible_appFun).

Lemma Lambda_redices_are_reductible:
  forall (G: LEnv) (t u: LTerm) (T U: LType),
    (WellTyped (snoc G T) t U) ->
    (Acc (symRel RedTerm) t) ->
    (Reductible G u T) ->
    (Reductible G (subst_in_term t 0 u) U) ->
    (Reductible G (LTerm_app (LTerm_fun T t) u) U)
.

Proof.
intros.
apply ProveReductible_appFun.PreRed_implies_Reductible.
unfold PreReductible_appFun.PreRed. eapply PreRed_appFun_intro; eauto.
Qed.

(************************************************************************)

Lemma SN_subst_inversion:
  forall (t u: LTerm),
    (Acc (symRel RedTerm) (subst_in_term t 0 u)) ->
    (Acc (symRel RedTerm) t)
.

Proof.
cut (
  forall (t': LTerm),
    (Acc (symRel RedTerm) t') ->
    forall (t u: LTerm),
      t' = (subst_in_term t 0 u) ->
      (Acc (symRel RedTerm) t)
).
intros. eapply H. apply H0. eauto.

intros t' H. induction H.
intros. apply Acc_intro. intros.
eapply H0. rewrite H1. unfold symRel. apply Substitution_in_reduction. apply H2.
eauto.
Qed.

Lemma SN_pair_intro:
  forall (t u: LTerm),
    (Acc (symRel RedTerm) t) ->
    (Acc (symRel RedTerm) u) ->
    (Acc (symRel RedTerm) (LTerm_pair t u))
.

Proof.
intros. generalize u H0. clear u H0. induction H.
intros. induction H1. apply Acc_intro. intros.
inversion H3.
eapply H0. trivial. apply Acc_intro; trivial.
eapply H2. trivial.
Qed.

Lemma SN_fun_intro:
  forall (t: LTerm) (T: LType),
    (Acc (symRel RedTerm) t) ->
    (Acc (symRel RedTerm) (LTerm_fun T t))
.

Proof.
intros. generalize T. clear T. induction H.
intros. apply Acc_intro. intros.
inversion H1.
eapply H0. trivial.
Qed.

(************************************************************************)

Theorem WellTyped_implies_Reductible_aux:
  forall (G1 Us : LEnv) (t : LTerm) (T : LType) (us : list LTerm),
    (WellTyped (G1 ++ Us) t T) ->
    (Forall2 (fun (u : LTerm) (U : LType) => (Reductible G1 u U)) us Us) ->
    (Reductible G1 (polySubst_in_term t 0 us) T)
.

Proof.
cut (
  forall (G : LEnv) (t : LTerm) (T : LType),
    (WellTyped G t T) ->
    forall (G1 G1' Us : LEnv) (us : list LTerm),
      G = (G1 ++ Us) ->
      (Forall2 (fun (u : LTerm) (U : LType) => (Reductible (G1' ++ G1) u U)) us Us) ->
      (Reductible (G1' ++ G1) (polySubst_in_term t 0 us) T)
).
intros. fold (nil ++ G1).
eapply H. apply H0. reflexivity. trivial.

intros G t T H. induction H.

(* WT-const *)
intros. rewrite polySubst_in_const.
apply Constants_are_reductible. apply WT_const; trivial.

(* WT-var *)
intros.
assert ((length us) = (length Us)). eapply Forall2_implies_same_length; eauto.
destruct (match_comp_nat2_aux_0 i (length us)).
(* i < |us| *)
destruct (List_from_domain _ _ _ l).
rewrite polySubst_in_var_variant with (u := x).
destruct (Forall2_and_ListFrom _ _ _ _ _ _ _ H1 H3). destruct H4.
rewrite H2 in l.
replace T with x0. trivial.
rewrite H0 in H.
eapply List_from_is_a_partial_function. apply H4.
eapply List_from_lem4. apply H. trivial.
trivial.
(* i >= |us| *)
rewrite polySubst_in_var1.
apply Variables_are_reductible.
apply WT_var.
rewrite H0 in H. rewrite H2.
apply List_from_lem1_cor1. rewrite <- H2. trivial.
replace ((G1' ++ G1) ++ Us) with (G1' ++ (G1 ++ Us)).
apply List_from_lem3; trivial. auto with datatypes. trivial.

(* WT-fun *)
assert (WellTyped G (LTerm_fun T t) (LType_fun T U)). apply WT_fun. trivial.
intros.
(* preliminary *)
set (v := (lift_in_term (LTerm_var 0) (length (G1' ++ G1)))).
assert (Reductible ((T :: G1') ++ G1) (polySubst_in_term t 0 (snoc us v)) U).
eapply IHWellTyped. rewrite H1. rewrite snoc_of_append. eauto.
apply Forall2_snoc.
eapply Forall2_implies with (P1 := fun u U => (Reductible (G1' ++ G1) u U)).
intros.
replace ((T :: G1') ++ G1) with ((T :: nil) ++ (G1' ++ G1)).
apply Reductible_left_weakening. trivial. auto with datatypes. trivial.
unfold v. unfold lift_in_term. simpl. apply Variables_are_reductible.
replace (LTerm_var (lift_at_depth_in_var 0 0 (length (G1' ++ G1))))
with (lift_in_term (LTerm_var 0) (length (G1' ++ G1))).
replace (T :: G1' ++ G1) with ((T :: nil) ++ (G1' ++ G1)).
apply Weakening_of_well_typing_by_the_right. apply WT_var.
replace (T :: nil) with (snoc nil T). apply List_from_tail. trivial.
auto with datatypes. trivial.

(* begin *)
rewrite polySubst_in_fun.
apply Reductible_fun.
(* *)
rewrite <- polySubst_in_fun.
rewrite H1 in H0.
eapply PolySubstitution_preserves_typing_cor1 with (Us := Us). 
replace ((G1' ++ G1) ++ Us) with (G1' ++ G1 ++ Us).
apply Weakening_of_well_typing_by_the_left. trivial. auto with datatypes.
apply Forall2_implies with (P1 := fun u U => (Reductible (G1' ++ G1) u U)).
apply Reductible_implies_WellTyped. trivial.
(* *)
apply SN_fun_intro.
eapply SN_subst_inversion.
rewrite polySubst_by_snoc in H3.
eauto with reductible_db.
(* *)
intros.
assert (Reductible (G' ++ G1' ++ G1) (subst_in_term (polySubst_in_term t 1 us) 0 u) U).
rewrite <- polySubst_by_snoc.
rewrite <- app_ass.
eapply IHWellTyped. rewrite H1. rewrite snoc_of_append. eauto.
apply Forall2_snoc.
rewrite app_ass.
apply Forall2_implies with (P1 := fun u U => (Reductible (G1' ++ G1) u U)).
intros. apply Reductible_left_weakening. trivial. trivial.
rewrite app_ass. trivial.
apply Lambda_redices_are_reductible.
rewrite H1 in H. rewrite snoc_of_append.
apply Weakening_of_well_typing_by_the_left.
unfold snoc. replace 1 with (length (T :: nil)).
eapply PolySubstitution_preserves_typing.
apply Forall2_implies with (P1 := fun u U => (Reductible (G1' ++ G1) u U)).
intros. eauto with reductible_db. apply H2.
rewrite app_ass. apply Weakening_of_well_typing_by_the_left.
rewrite snoc_of_append in H. unfold snoc in H. trivial. trivial.
eapply SN_subst_inversion.
eapply Reductible_implies_SN. apply H5. trivial. trivial.

(* WT-app *)
intros.
rewrite polySubst_in_app.
assert (Reductible (G1' ++ G1) (polySubst_in_term t 0 us) (LType_fun T U)).
eapply IHWellTyped1. apply H1. trivial.
assert (Reductible (G1' ++ G1) (polySubst_in_term u 0 us) T).
eapply IHWellTyped2. apply H1. trivial.
eauto with reductible_db.

(* WT-prod *)
intros.
assert (Reductible (G1' ++ G1) (polySubst_in_term t 0 us) T). eapply IHWellTyped1; eauto.
assert (Reductible (G1' ++ G1) (polySubst_in_term u 0 us) U). eapply IHWellTyped2; eauto.
rewrite polySubst_in_pair.
apply Reductible_prod.
apply WT_pair. eauto with reductible_db. eauto with reductible_db.
apply SN_pair_intro. eauto with reductible_db. eauto with reductible_db.
eapply First_projections_are_reductible.
trivial.
eauto with reductible_db.
eauto with reductible_db.
eapply Second_projections_are_reductible.
eauto with reductible_db.
eauto with reductible_db.
trivial.

(* WT-fst *)
intros. rewrite polySubst_in_fst.
assert (Reductible (G1' ++ G1) (polySubst_in_term t 0 us) (LType_prod T U)).
eapply IHWellTyped; eauto.
simpl in H2. tauto.

(* WT-snd *)
intros. rewrite polySubst_in_snd.
assert (Reductible (G1' ++ G1) (polySubst_in_term t 0 us) (LType_prod T U)).
eapply IHWellTyped; eauto.
simpl in H2. tauto.
Qed.

Theorem WellTyped_implies_Reductible:
  forall (G: LEnv) (t: LTerm) (T: LType),
    (WellTyped G t T) ->
    (Reductible G t T)
.

Proof.
intros.
replace t with (polySubst_in_term t 0 nil).
eapply WellTyped_implies_Reductible_aux with (Us := (nil (A := LType))).
rewrite <- app_nil_end. trivial.
apply Forall2_nil. trivial.
Qed.

Theorem WellTyped_implies_SN:
  forall (G: LEnv) (t: LTerm) (T: LType),
    (WellTyped G t T) ->
    (Acc (symRel RedTerm) t)
.

Proof.
intros.
eapply Reductible_implies_SN.
apply WellTyped_implies_Reductible. apply H.
Qed.

End LambdaCalculusNormalization.