(* Author: Vincent Cremet *)
(* Date: June 2007 *) 
(* File: SubjectReduction_preliminaries.v *)
(* Description: Subject reduction (preliminaries) *)

(** * Subject reduction (preliminaries) *)

Require Import List.
Require Import Arith.
Require Import Relations.

(* Require Import MatchCompNat. *)
Require Import Util.

Require Syntax.
Module SetProgram(My_Program: Syntax.Program).
Import My_Program.

(* Import definitions of TermSubstitutionInTyping.v *)
Require TermSubstitutionInTyping.
Module My_TermSubstitutionInTyping := TermSubstitutionInTyping.SetProgram(My_Program).
Import My_TermSubstitutionInTyping.

Import My_TypeSubstitutionInTyping.
Import My_BoundCompliantSubstitutions_proofs.
Import My_Progress_proofs.
Import My_TransitivityElimination_proofs.
Import My_Normalization_proofs.
Import My_Subtyping_proofs.
Import My_Typing_proofs.
Import My_Confluence.
Import My_Auxiliary_proofs.
Import My_Substitutions_proofs.
Import My_Binders_proofs.
Import My_Semantics.
Import My_WellFormedness.
Import My_Typing.
Import My_Auxiliary.
Import My_Substitutions.
Import My_Binders.

(********************************************************************)
(********************************************************************)

(** Beta-equivalence on types **)
Inductive BetaType: TypeS -> TypeS -> Prop :=

  | BetaType_step:
    forall (t u: TypeS),
      (RedType t u) ->
      (BetaType t u)

  | BetaType_refl:
    forall (t: TypeS),
      (BetaType t t)

  | BetaType_sym:
    forall (t u: TypeS),
      (BetaType t u) ->
      (BetaType u t)

  | BetaType_trans:
    forall (t1 t2 t3: TypeS),
      (BetaType t1 t2) ->
      (BetaType t2 t3) ->
      (BetaType t1 t3)
.

(**********************************************************************)

Lemma Class_type_subtyping_injection_aux1:
  forall (C C': ClassSym) (R R': TypeS),
    (Class_subtyping (Class_type C R) (Class_type C' R')) ->
    (Class_type C R) = (Class_type C' R') \/
    (Rplus classInheritance C C')
.

Proof.
cut (
  forall (N N': TypeS),
    (Class_subtyping N N') ->
    forall (C C': ClassSym) (R R': TypeS),
      N = (Class_type C R) ->
      N' = (Class_type C' R') ->
      (Class_type C R) = (Class_type C' R') \/
      (Rplus classInheritance C C')
).
intros. eapply H; eauto.

intros N N' H. induction H.
(* refl *)
intros. rewrite <- H.  rewrite <- H0. tauto.
(* extends *)
intros. inversion H1. clear H1. rewrite <- H4.  rewrite <- H5. clear C0 R0 H4 H5.
(* *)
destruct (classes_are_well_kinded C).
destruct (H3 _ H). clear H3 H5. destruct H4. destruct H3.
rename x into A. rename x0 into RA. rewrite H3 in H.
right.
(* *)
assert (classInheritance C A). eapply classInheritance_intro; eauto.
(* *)
assert ((typeSubst_in_type N 0 R) = (Class_type A (typeSubst_in_type RA 0 R))).
rewrite H3. rewrite typeSubst_of_class_type. trivial.
destruct (IHClass_subtyping _ _ _ _ H5 H2).
(* *)
inversion H6.
rewrite <- H8. apply tc_step; trivial.
(* *)
eapply tc_trans. apply tc_step. eauto. eauto.
Qed.

(**********************************************************************)

Lemma Accessibility_is_anti_reflexive:
  forall (A: Set) (R: A -> A -> Prop) (x: A),
    (Acc R x) ->
    ~(R x x)
.

Proof.
intros. induction H. intro. unfold not in H0. eapply H0.
apply H1. trivial.
Qed.

Lemma Class_inheritance_is_anti_reflexive:
  forall (C: ClassSym),
    ~(Rplus classInheritance C C)
.

Proof.
intros.
cut (~(symRel (Rplus classInheritance) C C)). tauto.
apply Accessibility_is_anti_reflexive with (R := (symRel (Rplus classInheritance))).
apply transitive_closure_well_founded.
apply Class_inheritance_is_well_founded.
Qed.

(**********************************************************************)

Lemma Class_type_subtyping_injection:
  forall (C: ClassSym) (R R': TypeS),
  (Class_subtyping (Class_type C R) (Class_type C R')) ->
  R = R'
.

Proof.
intros.
destruct (Class_type_subtyping_injection_aux1 _ _ _ _ H).
(* *)
inversion H0. trivial.
(* *)
destruct (Class_inheritance_is_anti_reflexive _ H0).
Qed.

(**********************************************************************)

Lemma Class_type_subtyping_is_linear:
  forall (N N1 N2: TypeS),
  (Class_subtyping N N1) ->
  (Class_subtyping N N2) ->
  (Class_subtyping N1 N2) \/ (Class_subtyping N2 N1)
.

Proof.
intros N N1 N2 H. generalize N2. clear N2. induction H.
(* refl *)
intros. tauto.
(* extends *)
intros N2 H'. inversion H'.
(* - refl *)
clear N2 H' C0 R0 H2 H3 H1.
right.
eapply CS_Extends; eauto.
(* - extends *)
rewrite H in H3. injection H3. intro. clear H3.
rewrite <- H6 in H5.
(* *)
auto.
Qed.

(**********************************************************************)

Theorem Uniqueness_of_class_type_subtyping:
  forall (A B: ClassSym) (R R1 R2: TypeS),
    (Class_subtyping (Class_type A R) (Class_type B R1)) ->
    (Class_subtyping (Class_type A R) (Class_type B R2)) ->
    R1 = R2
.

Proof.
intros.
destruct (Class_type_subtyping_is_linear _ _ _ H H0).
eapply Class_type_subtyping_injection; eauto.
symmetry. eapply Class_type_subtyping_injection; eauto.
Qed.

(**********************************************************************)

Lemma Class_subtyping_aux1:
  forall (C B: ClassSym) (R R' RB': TypeS),
    (RedTypeStar R R') ->
    (Class_subtyping (Class_type C R') (Class_type B RB')) ->
    exists RB: TypeS,
      (Class_subtyping (Class_type C R) (Class_type B RB)) /\
      (RedTypeStar RB RB')
.

Proof.
cut (
  forall (T U: TypeS),
    (Class_subtyping T U) ->
    forall (C B: ClassSym) (R R' RB': TypeS),
      T = (Class_type C R') ->
      U = (Class_type B RB') ->
      (RedTypeStar R R') ->
      exists RB: TypeS,
        (Class_subtyping (Class_type C R) (Class_type B RB)) /\
        (RedTypeStar RB RB')
).
intros. eapply H; eauto.

intros T N H. induction H.
(* refl *)
intros. inversion H.
exists R0. inversion H0. rewrite <- H3. rewrite <- H5.
rewrite <- H6. rewrite H4.
split. apply CS_Refl. trivial.
(* extends *)
intros.
(* *)
inversion H1. rewrite <- H5. clear C0 H1 H5. rewrite H6 in H0.
rewrite H6 in IHClass_subtyping. clear R H6.
rename R0 into R.
(* *)
destruct (classes_are_well_kinded C).
destruct (H4 _ H). clear H1 H4 H6. destruct H5. destruct H1.
rename x into C'. rename x0 into R0.
(* *)
assert ((typeSubst_in_type N 0 R') = (Class_type C' (typeSubst_in_type R0 0 R'))).
rewrite H1. rewrite typeSubst_of_class_type. trivial.
(* *)
assert (RedTypeStar (typeSubst_in_type R0 0 R) (typeSubst_in_type R0 0 R')).
apply Reduction_in_substitution_star; trivial.
(* *)
destruct (IHClass_subtyping _ _ _ _ _ H4 H2 H5). destruct H6. rename x into RB.
(* *)
exists RB. split.
eapply CS_Extends. eauto. rewrite H1. rewrite typeSubst_of_class_type. trivial.
trivial.
Qed.

(**********************************************************************)

Lemma RedTypeStar_implies_BetaType:
  forall (t t': TypeS),
    (RedTypeStar t t') ->
    (BetaType t t')
.

Proof.
intros. induction H.
apply BetaType_step; trivial.
apply BetaType_refl.
eapply BetaType_trans; eauto.
Qed.

(**********************************************************************)

Lemma SubtypingWTR_implies_class_type_subtyping:
  forall (D: TypeEnv) (C1 C2: ClassSym) (R1 R2: TypeS),
    (SubtypingWTR D (Class_type C1 R1) (Class_type C2 R2)) ->
    exists R2': TypeS,
      (Class_subtyping (Class_type C1 R1) (Class_type C2 R2')) /\
      (BetaType R2 R2')
.

Proof.
cut (
  forall (D: TypeEnv) (T1 T2: TypeS),
    (SubtypingWTR D T1 T2) ->
    forall (C1 C2: ClassSym) (R1 R2: TypeS),
      T1 = (Class_type C1 R1) ->
      T2 = (Class_type C2 R2) ->
      exists R2': TypeS,
        (Class_subtyping (Class_type C1 R1) (Class_type C2 R2')) /\
        (BetaType R2 R2')
).
intros. eapply H; eauto.

intros D T T' H. induction H.

(* S-Refl *)
intros. exists R2.
rewrite H in H0. inversion H0. rewrite <- H3. split.
apply CS_Refl. apply BetaType_refl.

(* S-Var *)
intros. discriminate H2.

(* S-Class *)
intros.
inversion H1. clear H1. rewrite H4 in H. rewrite H5 in H0. rewrite H5 in IHSubtypingWTR.
clear C R H4 H5.
(* *)
destruct (classes_are_well_kinded C1).
destruct (H3 _ H). clear Pi H1 H3 H5. destruct H4. destruct H1.
rename x into B. rename x0 into RB.
(* *)
assert ((typeSubst_in_type N 0 R1) = (Class_type B (typeSubst_in_type RB 0 R1))).
rewrite H1. rewrite typeSubst_of_class_type. trivial.
(* *)
destruct (IHSubtypingWTR _ _ _ _ H3 H2). destruct H4.
rename x into R2'.
(* *)
exists R2'. split.
eapply CS_Extends. apply H. rewrite H1. rewrite typeSubst_of_class_type. trivial.
trivial.

(* S-Beta-Left *)
intros. rewrite H1 in H. inversion H. inversion H6.
fold (Class_type C1 R') in H4. assert (T' = (Class_type C1 R')). symmetry; trivial.
destruct (IHSubtypingWTR _ _ _ _ H7 H2). destruct H8. rename x into R2'.
(* *)
assert (RedTypeStar R1 R'). apply RedType_step; trivial.
destruct (Class_subtyping_aux1 _ _ _ _ _ H10 H8). destruct H11. rename x into R2''.
(* *)
exists R2''. split. trivial.
apply BetaType_sym. eapply BetaType_trans.
apply RedTypeStar_implies_BetaType. apply H12. apply BetaType_sym; trivial.

(* S-Beta-Right *)
intros. rewrite H2 in H. inversion H. inversion H6.
fold (Class_type C2 R') in H4. assert (U' = (Class_type C2 R')). symmetry; trivial.
destruct (IHSubtypingWTR _ _ _ _ H1 H7). destruct H8. rename x into R2'.
exists R2'. split. trivial.
eapply BetaType_trans. apply BetaType_step. eauto. trivial.
Qed.

(**********************************************************************)

Theorem Uniqueness_of_subtyping:
  forall (A B: ClassSym) (R R1 R2: TypeS),
    (Type_WK nil (Class_type A R) Type_) ->
    (Type_WK nil (Class_type B R1) Type_) ->
    (Type_WK nil (Class_type B R2) Type_) ->
    (Subtyping nil (Class_type A R) (Class_type B R1)) ->
    (Subtyping nil (Class_type A R) (Class_type B R2)) ->
    (BetaType R1 R2)
.

Proof.
intros.
(* *)
assert (SubtypingWTR nil (Class_type A R) (Class_type B R1)).
apply TransitivityElimination; trivial. clear H2.
(* *)
assert (SubtypingWTR nil (Class_type A R) (Class_type B R2)).
apply TransitivityElimination; trivial. clear H3.
(* *)
clear H H0 H1.
(* *)
destruct (SubtypingWTR_implies_class_type_subtyping _ _ _ _ _ H4). clear H4.
rename x into R1'. destruct H.
(* *)
destruct (SubtypingWTR_implies_class_type_subtyping _ _ _ _ _ H2). clear H2.
rename x into R2'. destruct H1.
(* *)
replace R2' with R1' in H2.
(* *)
eapply BetaType_trans. apply H0. apply BetaType_sym. trivial.
(* *)
eapply Uniqueness_of_class_type_subtyping; eauto.
Qed.

(************************************************************************)

Theorem ChurchRosser:
  forall (t u: TypeS),
    (BetaType t u) ->
    exists s: TypeS,
      (RedTypeStar t s) /\
      (RedTypeStar u s)
.

Proof.
intros. induction H.
(* step *)
exists u. split. apply RedType_step; trivial. apply RedType_refl.
(* refl *)
exists t. split. apply RedType_refl. apply RedType_refl.
(* sym *)
destruct IHBetaType. rename x into s. exists s. tauto.
(* trans *)
destruct IHBetaType1. destruct H1. rename x into s1.
destruct IHBetaType2. destruct H3. rename x into s2.
destruct (Type_reduction_is_confluent _ _ _ H2 H3).
destruct H5. rename x into s.
exists s. split.
eapply RedType_trans; eauto.
eapply RedType_trans; eauto.
Qed.

(************************************************************************)

Lemma BetaType_implies_subtyping_lem1:
  forall (D: TypeEnv) (T R R': TypeS) (k: nat),
    (BetaType R R') ->
    (Subtyping D (typeSubst_in_type T k R) (typeSubst_in_type T k R'))
.

Proof.
intros.
destruct (ChurchRosser _ _ H). rename x into R0. destruct H0.
eapply S_beta_left_star.
apply Reduction_in_substitution_star. apply H0.
eapply S_beta_right_star. apply S_Refl.
apply Reduction_in_substitution_star. apply H1.
Qed.

(************************************************************************)

Lemma classTypeParams_closed:
  forall (C: ClassSym) (k: nat) (R: TypeS),
    let Pi := (classTypeParams C) in
    (typeSubst_in_type Pi k R) = Pi
.

Proof.
intros.
change k with ((length (nil (A := TypeS))) + k).
eapply Substitution_in_closed_type_aux.
destruct (classes_are_well_kinded C). apply H.
Qed.

Lemma methodTypeParams_closed:
  forall (m: MethodSym) (k: nat) (R: TypeS),
    let Pi := (methodTypeParams m) in
    (typeSubst_in_type Pi (S k) R) = Pi
.

Proof.
intros.
set (C := (methodOwner m)).
set (PiC := (classTypeParams C)).
change (S k) with ((length (PiC :: nil)) + k).
eapply Substitution_in_closed_type_aux.
destruct (method_declarations_are_well_formed m). eauto.
Qed.

Lemma methodType_closed:
  forall (m: MethodSym) (k: nat) (R: TypeS),
    let T := (methodType m) in
    (typeSubst_in_type T (S (S k)) R) = T
.

Proof.
intros.
set (C := (methodOwner m)).
set (PiC := (classTypeParams C)).
set (Pim := (methodTypeParams m)).
change (S (S k)) with ((length (PiC :: Pim :: nil)) + k).
eapply Substitution_in_closed_type_aux.
destruct (method_declarations_are_well_formed m). eauto.
Qed.

Lemma methodParams_closed:
  forall (m: MethodSym) (k: nat) (R: TypeS),
    let Ts := (methodParams m) in
    (typeSubst_in_termEnv Ts (S (S k)) R) = Ts
.

Proof.
intros.
set (C := (methodOwner m)).
set (PiC := (classTypeParams C)).
set (Pim := (methodTypeParams m)).
destruct (method_declarations_are_well_formed m).
unfold typeSubst_in_termEnv.
apply List_forall_implies_map_id with 
  (f := (fun T0 : TypeS => typeSubst_in_type T0 (S (S k)) R)).
apply List_forall_implies_forall with (P1 := fun U => Type_WK D U Type_).
intros.
change (S (S k)) with ((length D) + k).
eapply Substitution_in_closed_type_aux. eauto. trivial.
Qed.

(************************************************************************)

Inductive Conform (D: TypeEnv) (C: ClassSym) (R: TypeS): Prop :=

  | Conform_intro:
    (Type_WK D (Class_type C R) Type_) ->
    (* (SectionSat D R (lift_in_type (classTypeParams C) (length D))) -> *)
    (SectionSat D R (classTypeParams C)) ->
    (Conform D C R)
.

Theorem Class_subtyping_subject_reduction:
  forall (D: TypeEnv) (C C': ClassSym) (R R': TypeS),
    let N := (Class_type C R) in
    let N' := (Class_type C' R') in
    (TypeEnv_WK D) ->
    (Conform D C R) ->
    (Class_subtyping N N') ->
    (Conform D C' R')
.

Proof.
cut (
  forall (D: TypeEnv) (N N': TypeS),
    (TypeEnv_WK D) ->
    (Class_subtyping N N') ->
    forall (C C': ClassSym) (R R': TypeS),
      N  = (Class_type C R) ->
      N' = (Class_type C' R') ->
      (Conform D C R) ->
      (Conform D C' R')
).
intros. eapply H. trivial. apply H2. unfold N. eauto. unfold N'. eauto. trivial.

intros D N N' Hy H. induction H.

(* refl *)
intros. inversion H. inversion H0.
rewrite <- H5. rewrite H3. rewrite <- H6. rewrite H4. trivial.
(* extends *)
intros.
(* *)
inversion H1. clear H1. rewrite <- H5 in H3. rewrite <- H6 in H3. clear C0 R0 H5 H6.
(* *)
destruct (classes_are_well_kinded C).
destruct (H4 _ H). clear H4. destruct H6. destruct H5. destruct H5.
rename x into B. rename x0 into RB.
(* *)
assert (SectionSat (Pi :: nil) RB (classTypeParams B)).
apply H6. trivial.
(* *)
eapply IHClass_subtyping with (C := B).
rewrite H5. rewrite typeSubst_of_class_type. eauto. trivial.
(* *)
apply Conform_intro.
destruct H3.
eapply classtype_expansion_preserves_well_kindedness_1. apply H3.
rewrite <- H5. trivial.
(* *)
rewrite <- classTypeParams_closed with (k := 0) (R := R).
(* *)
destruct H3. inversion H3.
eapply Type_substitution_preserves_satisfaction_snoc.
apply H13. apply H8.
(* *)
apply WK_typeEnv_snoc. trivial.
replace D with (D ++ nil).
apply Weakening_of_well_kindedness_by_the_left.
destruct (classes_are_well_kinded C). apply H14.
rewrite <- app_nil_end. trivial.
(* *)
fold Pi. unfold snoc.
apply Satisfaction_left_weakening. trivial.
(* *)
replace (snoc D (classTypeParams C)) with ((snoc D (classTypeParams C)) ++ nil).
apply Weakening_of_well_kindedness_by_the_left.
destruct (classes_are_well_kinded B). apply H14.
rewrite <- app_nil_end. trivial.
Qed.

(*******************************************************************)



End SetProgram.
