(* Author: Vincent Cremet *)
(* Date: June 2007 *) 
(* File: Subtyping_proofs.v *)
(* Description: Properties of subtyping in FGJ-omega (Proofs) *)

(** * Properties of subtyping in FGJ-omega (Proofs) *)

Require Import List.
Require Import Arith.
Require Import Relations.

Require Import MatchCompNat.
Require Import Util.

Require Syntax.
Module SetProgram(My_Program: Syntax.Program).
Import My_Program.

(* Import definitions of Typing_proofs.v *)
Require Typing_proofs.
Module My_Typing_proofs := Typing_proofs.SetProgram(My_Program).
Import My_Typing_proofs.

Import My_Confluence.
Import My_Auxiliary_proofs.
Import My_Substitutions_proofs.
Import My_Binders_proofs.
Import My_Semantics.
Import My_WellFormedness.
Import My_Typing.
Import My_Auxiliary.
Import My_Substitutions.
Import My_Binders.

(********************************************************************)
(********************************************************************)

Lemma S_beta_right_star:
  forall (D : TypeEnv) (T U' U : TypeS),
    (Subtyping D T U') -> 
    (RedTypeStar U U') ->
    (Subtyping D T U)
.

Proof.
intros D T U' U H1 H2. induction H2.
(* Step *)
eapply S_Beta_Right.  apply H. trivial.
(* Refl *)
trivial.
(* Trans *)
tauto.
Qed.

Lemma S_beta_left_star:
  forall (D : TypeEnv) (T T' U : TypeS),
    (RedTypeStar T T') ->
    (Subtyping D T' U) -> 
    (Subtyping D T U)
.

Proof.
intros D T U' U H1 H2. induction H1.
(* Step *)
eapply S_Beta_Left.  apply H. trivial.
(* Refl *)
trivial.
(* Trans *)
tauto.
Qed.

Lemma RT_snoc_left:
  forall (D D': TypeEnv) (Pi: TypeS),
    (RedTypeEnv D D') ->
    (RedTypeEnv (snoc D Pi) (snoc D' Pi))
.

Proof.
intros.
induction H.
rewrite snoc_of_cons. rewrite snoc_of_cons. apply RTE_head; trivial.
rewrite snoc_of_cons. rewrite snoc_of_cons. apply RTE_tail; trivial.
Qed.

Lemma RT_snoc_right:
  forall (D: TypeEnv) (Pi Pi': TypeS),
    (RedType Pi Pi') ->
    (RedTypeEnv (snoc D Pi) (snoc D Pi'))
.

Proof.
intros. induction D.
unfold snoc. simpl. apply RTE_head; trivial.
rewrite snoc_of_cons. rewrite snoc_of_cons. apply RTE_tail; trivial.
Qed.

Lemma Reduction_preserves_kind_of:
  forall (t t': TypeS) (k: KindS),
    (kind_of t k) ->
    (RedType t t') ->
    (kind_of t' k)
.

Proof.
intros. generalize t' H0. clear t' H0. induction H.
intros. inversion H0.
apply Kind_of_section. apply IHkind_of. trivial.
intros. inversion H0.
apply Kind_of_param. apply IHkind_of. trivial.
apply Kind_of_param. trivial.
intros. inversion H0.
intros. inversion H1.
apply Kind_of_paramList_cons. apply IHkind_of1. trivial. trivial.
apply Kind_of_paramList_cons. trivial. apply IHkind_of2; trivial.
Qed.

Lemma Reduction_of_class_types:
  forall (C: ClassSym) (R T: TypeS),
    (RedType (Class_type C R) T) ->
    exists R': TypeS,
      T = (Class_type C R') /\
      (RedType R R') 
.

Proof.
intros.
unfold Class_type in H.
inversion H. inversion H3.
exists R'. tauto.
Qed.

Lemma Reduction_preserves_class_types:
  forall (T T': TypeS),
    (is_classType T) ->
    (RedType T T') ->
    (is_classType T')
.

Proof.
intros.
destruct H. destruct H.
rewrite H in H0. destruct (Reduction_of_class_types _ _ _ H0). destruct H1.
exists x. exists x1. trivial.
Qed.

Inductive clos_refl (A: Set) (R: A -> A -> Prop): A -> A -> Prop :=
  | r_step: forall (x y: A), (R x y) -> (clos_refl A R x y)
  | r_refl: forall (x: A), (clos_refl A R x x)
.

Inductive RedTypeEnv_snoc: TypeEnv -> TypeEnv -> Prop :=

  | RTE_snoc_tail:
    forall (D: TypeEnv) (Pi Pi': TypeS),
    (RedType Pi Pi') ->
    (RedTypeEnv_snoc (snoc D Pi) (snoc D Pi'))

  | RTE_snoc_head:
    forall (D D': TypeEnv) (Pi: TypeS),
    (RedTypeEnv_snoc D D') ->
    (RedTypeEnv_snoc (snoc D Pi) (snoc D' Pi))
.

Lemma RedTypeEnv_included_in_RedTypeEnv_snoc_aux1:
  forall (Pi Pi': TypeS) (D: TypeEnv),
    (RedType Pi Pi') ->
    (RedTypeEnv_snoc (Pi :: D) (Pi' :: D))
.

Proof.
intros. pattern D. apply (snoc_ind TypeS).
replace (Pi :: nil) with (snoc nil Pi). replace (Pi' :: nil) with (snoc nil Pi').
apply RTE_snoc_tail; trivial. trivial. trivial.
intros.  rewrite <- snoc_of_cons. rewrite <- snoc_of_cons.
apply RTE_snoc_head; trivial.
Qed.

Lemma RedTypeEnv_included_in_RedTypeEnv_snoc_aux2:
  forall (Pi: TypeS) (D D': TypeEnv),
    (RedTypeEnv_snoc D D') ->
    (RedTypeEnv_snoc (Pi :: D) (Pi :: D'))
.

Proof.
intros. induction H.
rewrite <- snoc_of_cons. rewrite <- snoc_of_cons.
apply RTE_snoc_tail; trivial.
rewrite <- snoc_of_cons. rewrite <- snoc_of_cons.
apply RTE_snoc_head; trivial.
Qed.

Lemma RedTypeEnv_included_in_RedTypeEnv_snoc:
  (SubRelation _ RedTypeEnv RedTypeEnv_snoc)
.

Proof.
unfold SubRelation. intros. induction H.
apply RedTypeEnv_included_in_RedTypeEnv_snoc_aux1; trivial.
apply RedTypeEnv_included_in_RedTypeEnv_snoc_aux2; trivial.
Qed.

Lemma RedTypeEnv_snoc_included_in_RedTypeEnv_aux1:
  forall (Pi Pi': TypeS) (D: TypeEnv),
    (RedType Pi Pi') ->
    (RedTypeEnv (snoc D Pi) (snoc D Pi'))
.

Proof.
intros. induction D.
unfold snoc. simpl.
apply RTE_head; trivial.
rewrite snoc_of_cons. rewrite snoc_of_cons.
apply RTE_tail; trivial.
Qed.

Lemma RedTypeEnv_snoc_included_in_RedTypeEnv_aux2:
  forall (Pi: TypeS) (D D': TypeEnv),
    (RedTypeEnv D D') ->
    (RedTypeEnv (snoc D Pi) (snoc D' Pi))
.

Proof.
intros. induction H.
rewrite snoc_of_cons. rewrite snoc_of_cons.
apply RTE_head; trivial.
rewrite snoc_of_cons. rewrite snoc_of_cons.
apply RTE_tail; trivial.
Qed.

Lemma RedTypeEnv_snoc_included_in_RedTypeEnv:
  (SubRelation _ RedTypeEnv_snoc RedTypeEnv)
.

Proof.
unfold SubRelation. intros. induction H.
apply RedTypeEnv_snoc_included_in_RedTypeEnv_aux1; trivial.
apply RedTypeEnv_snoc_included_in_RedTypeEnv_aux2; trivial.
Qed.

Lemma Reduction_preserves_TypeSection_at:
  forall (D D': TypeEnv) (n: nat) (Pi: TypeS),
    (TypeSection_at D n Pi) ->
    (RedTypeEnv D D') ->
    exists Pi': TypeS,
      (TypeSection_at D' n Pi') /\
      (clos_refl _ RedType Pi Pi')
.

Proof.
intros. generalize D' H0. clear D' H0. induction H.
(* TypeSection_at_tail *)
intros. assert (RedTypeEnv_snoc (snoc D Pi) D').
apply RedTypeEnv_included_in_RedTypeEnv_snoc; trivial. inversion H.
(* - RTE_snoc_tail *)
exists Pi'. split. apply TypeSection_at_tail. apply r_step.
replace Pi with Pi0. trivial. destruct (snoc_inject _ _ _ _ _ H1). trivial.
(* - RTE_snoc_head *)
exists Pi. split. replace Pi with Pi0. apply TypeSection_at_tail.
destruct (snoc_inject _ _ _ _ _ H1). trivial. apply r_refl.
(* TypeSection_at_head *)
intros. assert (RedTypeEnv_snoc (snoc D Pi) D').
apply RedTypeEnv_included_in_RedTypeEnv_snoc; trivial. inversion H1.
(* - RTE_snoc_tail *)
exists Pi'.  split. apply TypeSection_at_head. replace D0 with D. trivial.
destruct (snoc_inject _ _ _ _ _ H2). symmetry. trivial. apply r_refl.
(* - RTE_snoc_head *)
intros. assert (RedTypeEnv D D'0).
apply RedTypeEnv_snoc_included_in_RedTypeEnv; trivial. replace D with D0. trivial.
destruct (snoc_inject _ _ _ _ _ H2). trivial.
destruct (IHTypeSection_at _ H5). destruct H6.
exists x. split. apply TypeSection_at_head. trivial. trivial.
Qed.

Lemma Reduction_preserves_type_declaration_lookup:
  forall (D D': TypeEnv) (X: TypeVar) (Pi: TypeS),
    (decl_of_typeVar D X Pi) ->
    (RedTypeEnv D D') ->
    exists Pi': TypeS,
      (decl_of_typeVar D' X Pi') /\
      (clos_refl _ RedType Pi Pi')
.

Proof.
intros.
destruct H.
destruct (Reduction_preserves_TypeSection_at _ _ _ _ H H0). destruct H1.
destruct H2.
exists (lift_in_type y (S X)). split. apply decl_of_typeVar_intro. trivial.
apply r_step. unfold lift_in_type. apply Reduction_in_lifting; trivial.
exists (lift_in_type x (S X)). split. apply decl_of_typeVar_intro. trivial. apply r_refl.
Qed.

Lemma StarReduction_preserves_type_declaration_lookup:
  forall (D D': TypeEnv) (X: TypeVar) (Pi: TypeS),
    (decl_of_typeVar D X Pi) ->
    (RedTypeEnvStar D D') ->
    exists Pi': TypeS,
      (decl_of_typeVar D' X Pi') /\
      (RedTypeStar Pi Pi')
.

Proof.
intros. generalize Pi H. clear Pi H. induction H0.
(* step *)
intros. destruct (Reduction_preserves_type_declaration_lookup _ _ _ _ H0 H).
destruct H1. exists x. split. trivial. destruct H2. apply RedType_step; trivial.
apply RedType_refl.
(* refl *)
intros. exists Pi. split. trivial. apply RedType_refl.
(* trans *)
intros.
destruct (IHRedTypeEnvStar1 _ H). destruct H0.
destruct (IHRedTypeEnvStar2 _ H0). destruct H2.
exists x0. split. trivial. eapply RedType_trans. apply H1. trivial.
Qed.

Lemma Well_kindedness_is_preserved_by_type_env_reduction:
  forall (D D': TypeEnv) (t: TypeS) (c: TypeC),
    (Type_WK D t c) ->
    (RedTypeEnv D D') ->
    (Type_WK D' t c)
.

Proof.
intros. generalize D' H0. clear D' H0. induction H.
(* WK_proj *)
intros. eapply WK_proj. apply IHType_WK; trivial. trivial.
(* WK_class *)
intros. apply WK_class. trivial.
(* WK_fun *)
intros. apply WK_fun. apply IHType_WK1; trivial.
apply IHType_WK2. apply RT_snoc_left; trivial. trivial. trivial.
(* WK_apply *)
intros. eapply WK_apply. apply IHType_WK1; trivial. apply IHType_WK2; trivial.
(* WK_var *)
intros.
destruct (Reduction_preserves_type_declaration_lookup _ _ _ _ H H1). destruct H2.
eapply WK_var. apply H2. destruct H3.
eapply Reduction_preserves_kind_of. apply H0. trivial. trivial.
(* WK_tuple *)
intros. apply WK_tuple. apply IHType_WK; trivial.
(* WK_param *)
intros. apply WK_param. apply IHType_WK1; trivial.
apply IHType_WK2. apply RT_snoc_left; trivial. trivial. trivial.
(* WK_section *)
intros. apply WK_section. apply IHType_WK; trivial.
apply RT_snoc_left; trivial.
(* WK_none *)
intros. apply WK_none.
(* WK_some *)
intros. apply WK_some. trivial. apply IHType_WK; trivial.
(* WK_param_nil *)
intros. apply WK_param_nil.
(* WK_param_cons *)
intros. eapply WK_param_cons. apply IHType_WK1; trivial. apply IHType_WK2; trivial.
(* WK_constr_nil *)
intros. apply WK_constr_nil.
(* WK_constr_cons *)
intros. eapply WK_constr_cons. apply IHType_WK1; trivial. apply IHType_WK2; trivial.
Qed.

Lemma Well_kindedness_of_tuple:
  forall (D: TypeEnv) (Ks Ki: TypeS) (ks ki: KindS) (i: nat),
  (Type_WK D (Type_tuple Ks) (TypeTuple (Kind_tuple ks))) ->
  (TypeS_at Ks i Ki) ->
  (KindS_at ks i ki) ->
  (Type_WK D Ki (TypeConstr ki))
.

Proof.
intros. generalize ks ki H H1. clear ks ki H H1. induction H0.
(* TypeS_at_head *)
intros. inversion H1. inversion H. inversion H6.
replace ki with k1. trivial. rewrite <- H0 in H10. injection H10. intros.
replace ki with k. trivial.
(* TypeS_at_tail *)
intros. inversion H1. inversion H. inversion H9.
eapply IHTypeS_at. apply WK_tuple. apply H15. rewrite <- H3 in H13.
injection H13; intros. replace ks1 with ks0. trivial.
Qed.

Theorem Well_kindedness_is_preserved_by_type_reduction:
  forall (D: TypeEnv) (t t': TypeS) (c: TypeC),
    (Type_WK D t c) ->
    (RedType t t') ->
    (Type_WK D t' c)
.

Proof.
intros. generalize D c H. clear D c H. induction H0.
(* RT-proj-sub *)
intros. inversion H. eapply WK_proj. apply IHRedType. apply H4. trivial.
(* RT-proj-redex*)
intros. inversion H0. clear D0 i0 R H1 H3 H2. rename k into ks. rename k' into ki.
eapply Well_kindedness_of_tuple. apply H4. apply H. trivial.
(* RT-fun-left *)
intros. inversion H. apply WK_fun. apply IHRedType; trivial. trivial.
eapply Well_kindedness_is_preserved_by_type_env_reduction. apply H4.
apply RT_snoc_right; trivial. trivial.
eapply Reduction_preserves_kind_of. apply H8. trivial.
(* RT-fun-right *)
intros. inversion H. apply WK_fun. 
trivial.
apply IHRedType; trivial.
trivial. (* eapply Reduction_preserves_class_types. apply H6. trivial. *)
trivial. 
(* RT-apply-redex *)
intros. inversion H. inversion H3.
replace 0 with (length (nil (A := TypeS))).
replace D with (D ++ (typeSubst_in_typeEnv (nil (A := TypeS)) 0 R)).
eapply Type_substitution_preserves_well_kindedness.
apply H5. apply H13. apply H11. trivial. simpl. auto with datatypes. trivial.
(* RT-apply-left *)
intros. inversion H. eapply WK_apply. apply IHRedType; trivial. apply H4. trivial.
(* RT-apply-right *)
intros. inversion H. eapply WK_apply. apply H4. apply IHRedType; trivial.
(* RT-tuple *)
intros. inversion H. eapply WK_tuple. apply IHRedType; trivial.
(* RT-param-left *)
intros. inversion H. apply WK_param. apply IHRedType; trivial.
eapply Well_kindedness_is_preserved_by_type_env_reduction. apply H6.
apply RT_snoc_right; trivial.
(* RT-param-right *)
intros. inversion H. apply WK_param. trivial. apply IHRedType; trivial.
(* RT-section *)
intros. inversion H. eapply WK_section. apply IHRedType.
eapply Well_kindedness_is_preserved_by_type_env_reduction. apply H3.
apply RT_snoc_right. apply RT_section. trivial.
(* RT-some *)
intros. inversion H. eapply WK_some.
eapply Reduction_preserves_class_types. apply H2. trivial. apply IHRedType; trivial.
(* RT-cons-head *)
intros. inversion H.
apply WK_param_cons. apply IHRedType. trivial. trivial.
apply WK_constr_cons. apply IHRedType. trivial. trivial.
(* RT-cons-tail *)
intros. inversion H.
apply WK_param_cons. trivial. apply IHRedType. trivial.
apply WK_constr_cons. trivial. apply IHRedType. trivial.
Qed.

Lemma Well_kindedness_is_preserved_by_type_reduction_star:
  forall (D: TypeEnv) (t t': TypeS) (c: TypeC),
    (Type_WK D t c) ->
    (RedTypeStar t t') ->
    (Type_WK D t' c)
.

Proof.
intros. induction H0.
eapply Well_kindedness_is_preserved_by_type_reduction. apply H. trivial.
trivial. tauto.
Qed.

Lemma Well_kindedness_is_preserved_by_type_env_reduction_star:
  forall (D D': TypeEnv) (t: TypeS) (c: TypeC),
    (Type_WK D t c) ->
    (RedTypeEnvStar D D') ->
    (Type_WK D' t c)
.

Proof.
intros. induction H0.
eapply Well_kindedness_is_preserved_by_type_env_reduction. apply H. trivial.
trivial. tauto.
Qed.

Lemma Reduction_of_variable_types:
  forall (X: TypeVar) (i: nat) (R T: TypeS),
    (RedType (Var_type X i R) T) ->
    exists R': TypeS,
      T = (Var_type X i R') /\
      (RedType R R') 
.

Proof.
intros.
unfold Var_type in H.
inversion H. inversion H3. inversion H7.
exists R'. tauto.
Qed.

Lemma Reduction_of_variable_types_star:
  forall (X: TypeVar) (i: nat) (R T: TypeS),
    (RedTypeStar (Var_type X i R) T) ->
    exists R': TypeS,
      T = (Var_type X i R') /\
      (RedTypeStar R R') 
.

Proof.
cut (
  forall (X: TypeVar) (i: nat) (T U: TypeS),
    (RedTypeStar T U) ->
    forall (R: TypeS),
      T = (Var_type X i R) ->
      exists R': TypeS,
        U = (Var_type X i R') /\
        (RedTypeStar R R')
).
intros. eapply H. apply H0. trivial.
intros X i T U H. induction H.
(* step *)
intros. rewrite H0 in H. destruct (Reduction_of_variable_types _ _ _ _ H). destruct H1.
exists x. split. trivial. apply RedType_step; trivial.
(* refl *)
intros. exists R. split. trivial. apply RedType_refl.
(* trans *)
intros.
destruct (IHRedTypeStar1 _ H1). destruct H2.
destruct (IHRedTypeStar2 _ H2). destruct H4.
exists x0. split. trivial. eapply RedType_trans. apply H3. trivial.
Qed.

Lemma Reduction_of_class_types_star:
  forall (C: ClassSym) (R T: TypeS),
    (RedTypeStar (Class_type C R) T) ->
    exists R': TypeS,
      T = (Class_type C R') /\
      (RedTypeStar R R') 
.

Proof.
cut ( 
  forall (C: ClassSym) (T U: TypeS),
    (RedTypeStar T U) ->
    forall (R: TypeS),
    T = (Class_type C R) ->
    exists R': TypeS,
      U = (Class_type C R') /\
      (RedTypeStar R R')
).
intros. eapply H. apply H0. trivial.
intros C T U H. induction H.
(* step *)
intros. rewrite H0 in H. destruct (Reduction_of_class_types _ _ _ H). destruct H1.
exists x. split. trivial. apply RedType_step; trivial.
(* refl *)
intros. exists R. split. trivial. apply RedType_refl.
(* trans *)
intros.
destruct (IHRedTypeStar1 _ H1). destruct H2.
destruct (IHRedTypeStar2 _ H2). destruct H4.
exists x0. split. trivial. eapply RedType_trans. apply H3. trivial.
Qed.

Lemma RTS_section_inversion:
  forall (Ps Pi: TypeS),
    (RedTypeStar (Type_section Ps) Pi) ->
    exists Ps': TypeS,
      Pi = (Type_section Ps') /\
      (RedTypeStar Ps Ps')
.

Proof.
cut (
  forall (t Pi: TypeS),
    (RedTypeStar t Pi) ->
    forall (Ps: TypeS),
      t = (Type_section Ps) ->
      exists Ps': TypeS,
        Pi = (Type_section Ps') /\
        (RedTypeStar Ps Ps')
).
intros. eapply H.  apply H0. trivial.
intros t Pi H. induction H.
intros. rewrite H0 in H. inversion H. exists Ps'. split. trivial.
apply RedType_step; trivial.
intros. exists Ps. rewrite H. split. trivial. apply RedType_refl.
intros. destruct (IHRedTypeStar1 _ H1). destruct H2.
destruct (IHRedTypeStar2 _ H2). destruct H4. exists x0. split. trivial.
eapply RedType_trans. apply H3. trivial.
Qed.

Lemma Declaration_of_type_variable_is_preserved_by_reduction:
  forall (D D': TypeEnv) (X: TypeVar) (Ps: TypeS),
  (decl_of_typeVar D X (Type_section Ps)) ->
  (RedTypeEnvStar D D') ->
  exists Ps': TypeS,
    (decl_of_typeVar D' X (Type_section Ps')) /\
    (RedTypeStar Ps Ps')
.

Proof.
intros.
destruct (StarReduction_preserves_type_declaration_lookup _ _ _ _ H H0). destruct H1.
destruct (RTS_section_inversion _ _ H2). destruct H3. exists x0. split.
rewrite <- H3. trivial. trivial.
Qed.

Lemma Reduction_preserves_TypeS_at:
  forall (ts ts': TypeS) (i: nat) (t: TypeS),
    (TypeS_at ts i t) ->
    (RedType ts ts') ->
    exists t': TypeS,
      (TypeS_at ts' i t') /\
      (clos_refl _ RedType t t')
.

Proof.
intros. generalize ts' H0. clear ts' H0. induction H.
intros. inversion H0.
exists t'. split. apply TypeS_at_head. apply r_step; trivial.
exists t. split. apply TypeS_at_head. apply r_refl.
intros. inversion H0. 
exists u. split. apply TypeS_at_tail. trivial. apply r_refl.
clear t0 ts0 H1 H3. destruct (IHTypeS_at _ H4). destruct H1. exists x.
split. apply TypeS_at_tail. trivial. trivial.
Qed.

Lemma StarReduction_preserves_TypeS_at:
  forall (ts ts': TypeS) (i: nat) (t: TypeS),
    (TypeS_at ts i t) ->
    (RedTypeStar ts ts') ->
    exists t': TypeS,
      (TypeS_at ts' i t') /\
      (RedTypeStar t t')
.

Proof.
intros. generalize i t H. clear i t H. induction H0.
(* step *)
intros. destruct (Reduction_preserves_TypeS_at _ _ _ _ H0 H).
destruct H1. exists x. split. trivial. destruct H2. apply RedType_step; trivial.
apply RedType_refl.
(* refl *)
intros. exists t0. split. trivial. apply RedType_refl.
(* trans *)
intros.
destruct (IHRedTypeStar1 _ _ H). destruct H0.
destruct (IHRedTypeStar2 _ _ H0). destruct H2.
exists x0. split. trivial. eapply RedType_trans. apply H1. trivial.
Qed.

Lemma RTS_param_inversion:
  forall (Pi T P: TypeS),
    (RedTypeStar (Type_param Pi T) P) ->
    exists Pi': TypeS, exists T': TypeS,
      P = (Type_param Pi' T') /\
      (RedTypeStar Pi Pi') /\
      (RedTypeStar T T')
.

Proof.
cut (
  forall (t P: TypeS),
    (RedTypeStar t P) ->
    forall (Pi T: TypeS),
      t = (Type_param Pi T) ->
      exists Pi': TypeS, exists T': TypeS,
        P = (Type_param Pi' T') /\
        (RedTypeStar Pi Pi') /\
        (RedTypeStar T T')
).
intros. eapply H.  apply H0. trivial.
intros t P H. induction H.
intros. rewrite H0 in H. inversion H.
exists Pi'. exists T. split. trivial. split. apply RedType_step. trivial.
apply RedType_refl.
exists Pi. exists T'. split. trivial. split. apply RedType_refl.
apply RedType_step; trivial.
intros. exists Pi. exists T. rewrite H. split. trivial. split. apply RedType_refl.
apply RedType_refl.
intros. destruct (IHRedTypeStar1 _ _ H1). destruct H2. destruct H2. destruct H3.
destruct (IHRedTypeStar2 _ _ H2). destruct H5. destruct H5. destruct H6.
exists x1. exists x2. split. trivial. split.
eapply RedType_trans. apply H3. trivial.
eapply RedType_trans. apply H4. trivial.
Qed.

Lemma RTS_some_inversion:
  forall (Ps Pi: TypeS),
    (RedTypeStar (Type_some Ps) Pi) ->
    exists Ps': TypeS,
      Pi = (Type_some Ps') /\
      (RedTypeStar Ps Ps')
.

Proof.
cut (
  forall (t Pi: TypeS),
    (RedTypeStar t Pi) ->
    forall (Ps: TypeS),
      t = (Type_some Ps) ->
      exists Ps': TypeS,
        Pi = (Type_some Ps') /\
        (RedTypeStar Ps Ps')
).
intros. eapply H.  apply H0. trivial.
intros t Pi H. induction H.
intros. rewrite H0 in H. inversion H. exists T'. split. trivial.
apply RedType_step; trivial.
intros. exists Ps. rewrite H. split. trivial. apply RedType_refl.
intros. destruct (IHRedTypeStar1 _ H1). destruct H2.
destruct (IHRedTypeStar2 _ H2). destruct H4. exists x0. split. trivial.
eapply RedType_trans. apply H3. trivial.
Qed.

Lemma TypeS_at_is_preserved_by_reduction:
  forall (Ps Ps' Pi N: TypeS) (i: nat),
    (TypeS_at Ps i (Type_param Pi (Type_some N))) ->
    (RedTypeStar Ps Ps') ->
    exists Pi': TypeS, exists N': TypeS,
      (TypeS_at Ps' i (Type_param Pi' (Type_some N'))) /\
      (RedTypeStar Pi Pi') /\
      (RedTypeStar N N')
.

Proof.
intros.
destruct (StarReduction_preserves_TypeS_at _ _ _ _ H H0). destruct H1.
destruct (RTS_param_inversion _ _ _ H2). destruct H3. destruct H3. destruct H4.
destruct (RTS_some_inversion _ _ H5). destruct H6.
exists x0. exists x2. split.
rewrite H3 in H1. rewrite H6 in H1. trivial. split. trivial. trivial.
Qed.

(** Subtyping is preserved by type reduction *)
Lemma Reduction_preserves_subtyping:
  forall (D D': TypeEnv) (T U T' U': TypeS),
    (Subtyping D T U) ->
    (RedTypeEnvStar D D') ->
    (RedTypeStar T T') ->
    (RedTypeStar U U') ->
    (Subtyping D' T' U')
.

Proof.
intros D D' T U T' U' H. generalize D' T' U'. clear D' T' U'. induction H.
(* S-Refl *)
intros.
destruct (Type_reduction_is_confluent _ _ _ H0 H1). destruct H2. rename x into S.
eapply S_beta_right_star.
eapply S_beta_left_star. apply H2. apply S_Refl. trivial.
(* S-Trans *)
intros.
eapply S_Trans.
eapply Well_kindedness_is_preserved_by_type_env_reduction_star. apply H. trivial.
apply IHSubtyping1. trivial. trivial. apply RedType_refl.
apply IHSubtyping2. trivial. apply RedType_refl. trivial.
(* S-Var *)
intros.
destruct (Reduction_of_variable_types_star _ _ _ _ H3). destruct H5.
rename x into R'. rewrite H5.
destruct (Declaration_of_type_variable_is_preserved_by_reduction _ _ _ _ H H2).
destruct H7. rename x into Ps'.
assert (RedTypeStar (typeSubst_in_type Ps 0 (Type_var X))
                    (typeSubst_in_type Ps' 0 (Type_var X))).
apply Substitution_in_reduction_star. trivial. rename H9 into Hy.
destruct (TypeS_at_is_preserved_by_reduction _ _ _ _ _ H0 Hy). destruct H9.
destruct H9. destruct H10. 
eapply S_Var. apply H7. apply H9.
eapply IHSubtyping. trivial.
apply Reduction_and_substitution_star; trivial.
trivial.
(* S-Class *)
intros.
destruct (Reduction_of_class_types_star _ _ _ H2). destruct H4.
rename x into R'. rewrite H4.
eapply S_Class. apply H.
eapply IHSubtyping. trivial.
apply Reduction_in_substitution_star; trivial.
trivial.
(* S-Beta-Left *)
intros.
assert (RedTypeStar T T'). apply RedType_step; trivial.
destruct (Type_reduction_is_confluent _ _ _ H4 H2). destruct H5. rename x into S.
eapply S_beta_left_star. apply H6. apply IHSubtyping. trivial. trivial. trivial.
(* S-Beta-Right *)
intros.
assert (RedTypeStar U U'). apply RedType_step; trivial.
destruct (Type_reduction_is_confluent _ _ _ H4 H3). destruct H5. rename x into S.
eapply S_beta_right_star. apply IHSubtyping. trivial. trivial. apply H5. trivial.
Qed.

Lemma Reduction_preserves_removeBounds:
  forall (Pi Pi' x: TypeS),
    (removeBounds Pi x) ->
    (RedType Pi Pi') ->
    (removeBounds Pi' x)
.

Proof.
intros. generalize Pi' H0. clear Pi' H0. induction H.
intros. inversion H0.
apply removeBounds_param. apply IHremoveBounds. trivial.
apply removeBounds_param. trivial.
intros. inversion H0. apply removeBounds_section. apply IHremoveBounds; trivial.
intros. inversion H0.
intros. inversion H1.
apply removeBounds_param_cons. apply IHremoveBounds1; trivial. trivial.
apply removeBounds_param_cons. trivial. apply IHremoveBounds2; trivial.
Qed.

Lemma StarReduction_preserves_removeBounds:
  forall (Pi Pi' x: TypeS),
    (removeBounds Pi x) ->
    (RedTypeStar Pi Pi') ->
    (removeBounds Pi' x)
.

Proof.
intros. induction H0.
eapply Reduction_preserves_removeBounds. apply H. trivial. trivial. tauto.
Qed.

Lemma RTS_snoc_left:
  forall (D D' : TypeEnv) (Pi : TypeS),
    (RedTypeEnvStar D D') ->
    (RedTypeEnvStar (snoc D Pi) (snoc D' Pi))
.

Proof.
intros. induction H.
apply RedTypeEnv_step. apply RT_snoc_left; trivial.
apply RedTypeEnv_refl.
eapply RedTypeEnv_trans. apply IHRedTypeEnvStar1. trivial.
Qed.

Lemma RTS_snoc_right:
  forall (D: TypeEnv) (Pi Pi' : TypeS),
    (RedTypeStar Pi Pi') ->
    (RedTypeEnvStar (snoc D Pi) (snoc D Pi'))
.

Proof.
intros. induction H.
apply RedTypeEnv_step. apply RT_snoc_right; trivial.
apply RedTypeEnv_refl.
eapply RedTypeEnv_trans. apply IHRedTypeStar1. trivial.
Qed.

Lemma RTS_some_inversion_right:
  forall (D: TypeEnv) (t t1': TypeS),
    (Type_WK D t TypeOption) ->
    (RedTypeStar t (Type_some t1')) ->
    exists t1: TypeS,
      t = (Type_some t1) /\
      (RedTypeStar t1 t1')
.

Proof.
cut (
  forall (t t': TypeS),
    (RedTypeStar t t') ->
    forall (D: TypeEnv) (t1': TypeS),
      (Type_WK D t TypeOption) ->
      t' = (Type_some t1') ->
      exists t1: TypeS,
        t = (Type_some t1) /\
        (RedTypeStar t1 t1')
).
intros. eapply H.  apply H1. apply H0. trivial.
intros t t' H. induction H.
(* step *)
intros. rewrite H1 in H. inversion H.
rewrite <- H3 in H0. inversion H0.
rewrite <- H2 in H0. inversion H0.
exists T. split. trivial. apply RedType_step; trivial.
(* refl *)
intros. exists t1'. split. trivial. apply RedType_refl.
(* trans *)
intros.
assert (Type_WK D u TypeOption).
eapply Well_kindedness_is_preserved_by_type_reduction_star. apply H1. trivial.
destruct (IHRedTypeStar2 _ _ H3 H2).  destruct H4.
destruct (IHRedTypeStar1 _ _ H1 H4).  destruct H6.
exists x0. split. trivial. eapply RedType_trans. apply H7. trivial.
Qed.

(*******************************************************************)

Lemma RedTypeParTrans_in_lifting:
  forall (t t': TypeS) (n: nat),
    (RedTypeParTrans t t') ->
    forall (k: nat),
      (RedTypeParTrans (lift_at_depth_in_type t n k) (lift_at_depth_in_type t' n k))
.

Proof.
intros. induction H.
unfold RedTypeParTrans. apply t_step. apply Parallel_reduction_in_lifting. trivial.
unfold RedTypeParTrans. eapply t_trans; eauto.
Qed.

Lemma RedTypeStar_in_lifting:
  forall (t t': TypeS) (n: nat),
    (RedTypeStar t t') ->
    forall (k: nat),
      (RedTypeStar (lift_at_depth_in_type t n k) (lift_at_depth_in_type t' n k))
.

Proof.
intros.
apply RedTypeParTrans_included_in_RedTypeStar.
apply RedTypeParTrans_in_lifting.
apply RedTypeStar_included_in_RedTypeParTrans. trivial.
Qed.

(******************************************************************)

(** Type reduction preserves satisfaction *)
Theorem Reduction_preserves_satisfaction:
  forall (D D': TypeEnv) (K P K' P': TypeS),
    (Type_WK D P TypeParam) ->
    (ParamSat D K P) ->
    (RedTypeEnvStar D D') ->
    (RedTypeStar K K') ->
    (RedTypeStar P P') ->
    (ParamSat D' K' P')
.

Proof.
intros. destruct H0.
destruct (RTS_param_inversion _ _ _ H3). destruct H5. destruct H5. destruct H6.
rewrite H5. eapply ParamSat_intro. eapply StarReduction_preserves_removeBounds.
apply H0. trivial.
intros. rewrite H8 in H7.
inversion H. clear D0 Pi0 N_opt0 H11 H9 H10.
destruct (RTS_some_inversion_right _ _ _ H13 H7). destruct H9.
eapply Reduction_preserves_subtyping.
eapply H4. apply H9.
apply RTS_snoc_left; trivial.
apply RTS_apply_left.
unfold lift_in_type. apply RedTypeStar_in_lifting. trivial.
trivial.
Qed.

Lemma RTS_nil_inversion:
  forall (t': TypeS),
    (RedTypeStar Type_nil t') ->
    t' = Type_nil
.

Proof.
cut (
  forall (t t': TypeS),
    (RedTypeStar t t') ->
    t = Type_nil ->
    t' = Type_nil
).
intros. eapply H. apply H0. trivial.
intros t t' H. induction H.
intro. rewrite H0 in H. inversion H.
intro. trivial.
intro. tauto.
Qed.

Lemma RTS_cons_inversion:
  forall (t ts us: TypeS),
    (RedTypeStar (Type_cons t ts) us) ->
    exists t': TypeS, exists ts': TypeS,
      us = (Type_cons t' ts') /\
      (RedTypeStar t t') /\
      (RedTypeStar ts ts')
.

Proof.
cut (
  forall (t1 us: TypeS),
    (RedTypeStar t1 us) ->
    forall (t ts: TypeS),
      t1 = (Type_cons t ts) ->
      exists t': TypeS, exists ts': TypeS,
        us = (Type_cons t' ts') /\
        (RedTypeStar t t') /\
        (RedTypeStar ts ts')
).
intros. eapply H.  apply H0. trivial.
intros t1 us H. induction H.
intros. rewrite H0 in H. inversion H.
exists t'0. exists ts. split. trivial. split. apply RedType_step. trivial.
apply RedType_refl.
exists t0. exists ts'. split. trivial. split. apply RedType_refl.
apply RedType_step; trivial.
intros. exists t0. exists ts. rewrite H. split. trivial. split. apply RedType_refl.
apply RedType_refl.
intros. destruct (IHRedTypeStar1 _ _ H1). destruct H2. destruct H2. destruct H3.
destruct (IHRedTypeStar2 _ _ H2). destruct H5. destruct H5. destruct H6.
exists x1. exists x2. split. trivial. split.
eapply RedType_trans. apply H3. trivial.
eapply RedType_trans. apply H4. trivial.
Qed.

Lemma RTS_cons_inversion_right:
  forall (D: TypeEnv) (Ps Ps' P': TypeS) (i: nat),
    (Type_WK D Ps TypeParamList) ->
    (RedTypeStar Ps Ps') ->
    (TypeS_at Ps' i P') ->
    exists P: TypeS,
      (TypeS_at Ps i P) /\
      (RedTypeStar P P')
.

Proof.
intros. generalize Ps H H0. clear Ps H H0. induction H1.
(* *)
intros. inversion H.
rewrite <- H2 in H0.
absurd (RedTypeStar Type_nil (Type_cons t ts)). intro.
assert ((Type_cons t ts) = Type_nil). apply RTS_nil_inversion; trivial. discriminate H4.
trivial.
rewrite <- H4 in H0.
destruct (RTS_cons_inversion _ _ _ H0). destruct H5. destruct H5. destruct H6.
exists P. split. apply TypeS_at_head. injection H5. intros. rewrite H9. trivial.
(* *)
intros. inversion H.
rewrite <- H3 in H0.
absurd (RedTypeStar Type_nil (Type_cons t ts)). intro.
assert ((Type_cons t ts) = Type_nil). apply RTS_nil_inversion; trivial. discriminate H5.
trivial.
rewrite <- H5 in H0.
destruct (RTS_cons_inversion _ _ _ H0). destruct H6. destruct H6. destruct H7.
injection H6. intros. rewrite <- H9 in H8.
destruct (IHTypeS_at _ H3 H8). destruct H11. exists x1. split.
apply TypeS_at_tail. trivial. trivial.
Qed.

(********************************************************************)

Lemma TypeParamList_WK_inversion:
  forall (D: TypeEnv) (i: nat) (Ps P: TypeS),
    (Type_WK D Ps TypeParamList) ->
    (TypeS_at Ps i P) ->
    (Type_WK D P TypeParam)
.

Proof.
intros. generalize H. clear H. induction H0.
intros. inversion H. trivial.
intros. inversion H. tauto.
Qed.

(********************************************************************)

(** Type reduction preserves forall satisfaction *)
Theorem Reduction_preserves_forall_satisfaction:
  forall (D D': TypeEnv) (R Ps R' Ps': TypeS),
    (Type_WK D Ps TypeParamList) ->
    (Forall_ParamSat D R Ps) ->
    (RedTypeEnvStar D D') ->
    (RedTypeStar R R') ->
    (RedTypeStar Ps Ps') ->
    (Forall_ParamSat D' R' Ps')
.

Proof.
intros. unfold Forall_ParamSat. intros. rename P into P'.
unfold Forall_ParamSat in H0.
destruct (RTS_cons_inversion_right _ _ _ _ _ H H3 H4). rename x into P. destruct H5.
eapply Reduction_preserves_satisfaction. eapply TypeParamList_WK_inversion. apply H. apply H5.
apply H0. apply H5. trivial. apply RTS_proj_sub; trivial. trivial.
Qed.

(** Type reduction preserves forall satisfaction *)
Theorem Reduction_preserves_forall_satisfaction_build:
  forall (D D': TypeEnv) (i: nat) (R Ps R' Ps': TypeS),
    (Type_WK D Ps TypeParamList) ->
    (Forall_ParamSat_build D R i Ps) ->
    (RedTypeEnvStar D D') ->
    (RedTypeStar R R') ->
    (RedTypeStar Ps Ps') ->
    (Forall_ParamSat_build D' R' i Ps')
.

Proof.
intros. generalize D' R' Ps' H H1 H2 H3. clear D' R' Ps' H H1 H2 H3. induction H0.
(* nil *)
intros.
rewrite (RTS_nil_inversion _ H3). apply Forall_ParamSat_build_nil.
(* cons *)
intros.
destruct (RTS_cons_inversion _ _ _ H4). destruct H5. destruct H5. destruct H6.
rewrite H5.
inversion H1.
apply Forall_ParamSat_build_cons.
eapply Reduction_preserves_satisfaction. apply H11.
apply H. trivial. apply RTS_proj_sub. trivial. trivial.
apply IHForall_ParamSat_build. trivial. trivial. trivial. trivial.
Qed.

Lemma StarReduction_preserves_kind_of:
   forall (t t' : TypeS) (k : KindS),
     (kind_of t k) ->
     (RedTypeStar t t') ->
     (kind_of t' k)
.

Proof.
intros. induction H0.
eapply Reduction_preserves_kind_of. apply H. trivial. trivial. tauto.
Qed.

(** Type reduction preserves simultaneous satisfaction *)
Theorem Reduction_preserves_simultaneous_satisfaction:
  forall (D D': TypeEnv) (R R' Pi Pi': TypeS),
    (Type_WK D Pi TypeSection) ->
    (SectionSat D R Pi) ->
    (RedTypeEnvStar D D') ->
    (RedTypeStar R R') ->
    (RedTypeStar Pi Pi') ->
    (SectionSat D' R' Pi')
.

Proof.
intros. inversion H0.
rewrite <- H7 in H3.
destruct (RTS_section_inversion _ _ H3). destruct H8. rewrite H8.
rename x into Ps'.
eapply SectionSat_intro.
eapply StarReduction_preserves_kind_of. apply H4. trivial.
eapply Well_kindedness_is_preserved_by_type_env_reduction_star.
eapply Well_kindedness_is_preserved_by_type_reduction_star. apply H5.
trivial. trivial.
rewrite <- H7 in H. inversion H.
eapply Reduction_preserves_forall_satisfaction with (Ps := (typeSubst_in_type Ps 0 R)).
eapply Substitution_preserves_well_kindedness_snoc. apply H5.
apply Kind_of_section. apply H4.
apply H12. apply H6. trivial. trivial.
apply Reduction_and_substitution_star. trivial. trivial.
Qed.

(************************************************************************)

Lemma Subtyping_left_weakening:
  forall (D D': TypeEnv) (T U: TypeS),
    (Subtyping D T U) ->
    (Subtyping (D' ++ D) T U)
.

Proof.
intros. induction H.
(* S-refl *)
apply S_Refl.
(* S-trans *)
eapply S_Trans; eauto. apply Weakening_of_well_kindedness_by_the_left; trivial.
(* S-Var *)
eapply S_Var; eauto. apply Weakening_of_type_declaration_by_the_left; trivial.
(* S-Class *)
eapply S_Class; eauto.
(* S-beta-left *)
eapply S_Beta_Left; eauto.
(* S-beta-right *)
eapply S_Beta_Right; eauto.
Qed.

Lemma Param_satisfaction_left_weakening:
  forall (D D': TypeEnv) (K P: TypeS),
    (ParamSat D K P) ->
    (ParamSat (D' ++ D) K P)
.

Proof.
intros. inversion H.
eapply ParamSat_intro. apply H0. intros.
rewrite snoc_of_append.
apply Subtyping_left_weakening. auto.
Qed.

Lemma Satisfaction_left_weakening:
  forall (D D': TypeEnv) (R Pi: TypeS),
    (SectionSat D R Pi) ->
    (SectionSat (D' ++ D) R Pi)
.

Proof.
intros.
inversion H.
eapply SectionSat_intro. apply H0.
apply Weakening_of_well_kindedness_by_the_left; trivial.
unfold Forall_ParamSat. intros.
unfold Forall_ParamSat in H2.
apply Param_satisfaction_left_weakening. auto.
Qed.

(********************************************************************)

Lemma TypeSection_at_lem9:
  forall (D: TypeEnv) (i: nat) (Pi: TypeS),
    (TypeSection_at D i Pi) ->
    i < (length D)
.

Proof.
intros.
induction H. rewrite length_of_snoc. omega. rewrite length_of_snoc. omega.
Qed.

Lemma Lifting_in_closed_type_aux:
  forall (D: TypeEnv) (t: TypeS) (c: TypeC) (k n: nat),
    (Type_WK D t c) ->
    (lift_at_depth_in_type t n ((length D) + k)) = t
.

Proof.
intros. generalize k. clear k. induction H.
(* WK-proj *)
intro. simpl. rewrite IHType_WK. trivial.
(* WK-class *)
intros. simpl. trivial.
(* WK-fun *)
intro. simpl. rewrite IHType_WK1.
rewrite length_of_snoc in IHType_WK2. simpl in IHType_WK2. rewrite IHType_WK2. trivial.
(* WK-apply *)
intro. simpl. rewrite IHType_WK1. rewrite IHType_WK2. trivial.
(* WK-var *)
intro. simpl. destruct H. unfold lift_at_depth_in_typeVar.
rewrite match_comp_nat2_lt. trivial.
cut (X < (length D)). intro. omega. eapply TypeSection_at_lem9. apply H.
(* WK-tuple *)
intro. simpl. rewrite IHType_WK. trivial.
(* WK-param *)
intro. simpl. rewrite IHType_WK1.
rewrite length_of_snoc in IHType_WK2. simpl in IHType_WK2. rewrite IHType_WK2. trivial.
(* WK-section *)
intro. simpl.
rewrite length_of_snoc in IHType_WK. simpl in IHType_WK. rewrite IHType_WK. trivial.
(* WK-none *)
intro. simpl. trivial.
(* WK-some *)
intro. simpl. rewrite IHType_WK. trivial.
(* WK-param-nil *)
intro. simpl. trivial.
(* WK-param-cons *)
intro. simpl. rewrite IHType_WK1. rewrite IHType_WK2. trivial.
(* WK-constr-nil *)
intro. simpl. trivial.
(* WK-constr-cons *)
intro. simpl. rewrite IHType_WK1. rewrite IHType_WK2. trivial.
Qed.

Lemma Lifting_in_closed_type_cor1:
  forall (t: TypeS) (c: TypeC) (k n: nat),
    (Type_WK nil t c) ->
    (lift_at_depth_in_type t n k) = t
.

Proof.
cut (
  forall (t: TypeS) (c: TypeC) (k n: nat),
    (Type_WK nil t c) ->
    (lift_at_depth_in_type t n ((length (nil (A := TypeS))) + k)) = t
).
intros. simpl in H. eapply H; eauto.
intros. eapply Lifting_in_closed_type_aux; eauto.
Qed.

Lemma Lifting_in_closed_type_cor2:
  forall (Pi t: TypeS) (c: TypeC) (k n: nat),
    (Type_WK (Pi :: nil) t c) ->
    (lift_at_depth_in_type t n (S k)) = t
.

Proof.
intros.
cut (
  (lift_at_depth_in_type t n ((length (Pi :: nil)) + k)) = t
).
tauto.
eapply Lifting_in_closed_type_aux; eauto.
Qed.

(************************************************************************)

Lemma Weakening_of_subtyping:
  forall (D1 D2 D3: TypeEnv) (T U: TypeS),
    (Subtyping (D1 ++ D3) T U) ->
    (Subtyping (D1 ++ D2 ++ (lift_in_typeEnv D3 (length D2)))
               (lift_at_depth_in_type T (length D2) (length D3))
               (lift_at_depth_in_type U (length D2) (length D3)))
.

Proof.
cut (
  forall (D1 D2 D: TypeEnv) (T U: TypeS),
    (Subtyping D T U) ->
    forall (D3: TypeEnv),
      D = (D1 ++ D3) ->
      (Subtyping (D1 ++ D2 ++ (lift_in_typeEnv D3 (length D2)))
                 (lift_at_depth_in_type T (length D2) (length D3))
                 (lift_at_depth_in_type U (length D2) (length D3)))
).
intros. eapply H; eauto.
intros D1 D2 D T U H. induction H.

(* S-Refl *)
intros; simpl; subst D; econstructor; eauto.
(* S-Trans *)
intros; simpl; subst D; econstructor; eauto.
apply Weakening_of_well_kindedness; auto.
(* S-Var *)
intros. subst D.
change  (lift_at_depth_in_type (Var_type X i R) (length D2) (length D3)) with
        (Var_type (lift_at_depth_in_typeVar X (length D2) (length D3)) i
                  (lift_at_depth_in_type R (length D2) (length D3))).
eapply S_Var with 
  (Ps := (lift_at_depth_in_type Ps (length D2) (S (length D3))))
  (Pi := (lift_at_depth_in_type Pi (length D2) (length D3)))
  (N := (lift_at_depth_in_type N (length D2) (S (length D3))))
.
change (Type_section (lift_at_depth_in_type Ps (length D2) (S (length D3)))) with
       (lift_at_depth_in_type (Type_section Ps) (length D2) (length D3)).
apply Weakening_of_type_declaration; trivial.
change (Type_var (lift_at_depth_in_typeVar X (length D2) (length D3))) with
       (lift_at_depth_in_type(Type_var X) (length D2) (length D3)).
rewrite <- Permuting_type_lifting_and_substitution_2_cor1.
change (Type_param (lift_at_depth_in_type Pi (length D2) (length D3))
         (Type_some (lift_at_depth_in_type N (length D2) (S (length D3))))) with
     (lift_at_depth_in_type (Type_param Pi (Type_some N)) (length D2) (length D3)).
apply TypeS_at_of_lift; trivial.
rewrite <- Permuting_type_lifting_and_substitution_2_cor1. auto.
(* S-Class *)
intros. subst D.
change (lift_at_depth_in_type (Class_type C R) (length D2) (length D3)) with
 (Class_type C (lift_at_depth_in_type R (length D2) (length D3))).
eapply S_Class. eauto.
replace N with (lift_at_depth_in_type N (length D2) (S (length D3))).
rewrite <- Permuting_type_lifting_and_substitution_2_cor1. auto.
destruct (classes_are_well_kinded C). destruct (H2 _ H). destruct H4.
eapply Lifting_in_closed_type_cor2; eauto.
(* S-Beta-Left *)
intros. subst D. eapply S_Beta_Left. apply Reduction_in_lifting; eauto. auto.
(* S-Beta-Right *)
intros. subst D. eapply S_Beta_Right. apply Reduction_in_lifting; eauto. auto.
Qed.

End SetProgram.