(* Author: Vincent Cremet *)
(* Date: December 2006 *) 
(* File: Syntax.v *)
(* Description: Syntax of FGJ-omega (same for FGJ-Omega) *)

(** * Definition of Syntax *)

Require Import List.

(** Programs **)
Module Type Program.

  (** Symbols **)
  Parameter  ClassSym   : Set.              (* class symbols  -- C *)
  Parameter  MethodSym  : Set.              (* method symbols -- m *)
  Parameter  FieldSym   : Set.              (* field symbols  -- f *)

  (** Variables **)
  Definition Var        : Set := nat.       (* variables             -- x *)
  Definition TypeVar    : Set := nat.       (* typle tuple variables -- X *)

  (** Decidability of equality for all kinds of symbols **)
  Axiom ClassSym_eqdec  : forall (C C': ClassSym),  { C = C' } + { C <> C' }.
  Axiom FieldSym_eqdec  : forall (f f': FieldSym),  { f = f' } + { f <> f' }.
  Axiom MethodSym_eqdec : forall (m m': MethodSym), { m = m' } + { m <> m' }.

  (** Type syntax **)
  Inductive TypeS: Set :=
  (** Type constructors -- K **)
    | Type_proj    : TypeS -> nat -> TypeS      (* R@i *)
    | Type_class   : ClassSym -> TypeS          (* C *)
    | Type_fun     : TypeS -> TypeS -> TypeS    (* Pi => T *)
  (** Types -- T **)
    | Type_apply   : TypeS -> TypeS -> TypeS    (* K<R> *)
  (** Tuples of type constructors -- R **)
    | Type_var     : TypeVar -> TypeS           (* X *)
    | Type_tuple   : TypeS -> TypeS             (* (Ks) *)
  (** Type constructor parameters (TCP) -- P **)
    | Type_param   : TypeS -> TypeS -> TypeS    (* <Pi> extends T_opt *)
  (** Sections of type constructor parameters -- Pi **)
    | Type_section : TypeS -> TypeS             (* <Ps> *)
  (** Optional types -- T_opt **)
    | Type_none    : TypeS
    | Type_some    : TypeS -> TypeS
  (** Lists of TCPs and type constructors -- Ps and Ks **)
    | Type_nil    : TypeS
    | Type_cons   : TypeS -> TypeS -> TypeS 
  .

(*   (\** Kinds -- k_t **\) *)
(*   Inductive TupleKind: Set := *)
(*     | Kind_tuple : ConstrKindList -> TupleKind   (\* ( k_f's ) *\) *)

(*   (\** Kinds -- k_f **\) *)
(*   with ConstrKind: Set := *)
(*     | Kind_constr : TupleKind -> ConstrKind          (\* < k_t > *\) *)

(*   (\** Lists of functional kinds -- k_f's **\) *)
(*   with ConstrKindList: Set := *)
(*     | Kind_nil  : ConstrKindList *)
(*     | Kind_cons : ConstrKind -> ConstrKindList -> ConstrKindList *)
(*   . *)

  (** Kinds -- k **)
  Inductive KindS: Set :=
    | Kind_tuple  : KindS -> KindS           (* k_t ::= ( k_f's ) *)
    | Kind_constr : KindS -> KindS           (* k_f ::= < k_t > *)
    | Kind_nil    : KindS
    | Kind_cons   : KindS -> KindS -> KindS  (* k_f :: k_f's *)
  .

  (** Kinds categories -- c **)
  Inductive KindC: Set :=
    | TupleKind      : KindC
    | ConstrKind     : KindC
    | ConstrKindList : KindC
  .

  Inductive Kind_WD: KindS -> KindC -> Prop :=

    | WD_tuple:
      forall (k: KindS),
        (Kind_WD k ConstrKindList) ->
        (Kind_WD (Kind_tuple k) TupleKind)

    | WD_constr:
      forall (k: KindS),
        (Kind_WD k TupleKind) ->
        (Kind_WD (Kind_constr k) ConstrKind)

    | WD_nil:
      (Kind_WD Kind_nil ConstrKindList)

    | WD_cons:
      forall (k ks: KindS),
        (Kind_WD k ConstrKind) ->
        (Kind_WD ks ConstrKindList) ->
        (Kind_WD (Kind_cons k ks) ConstrKindList)
   .

  (** Type categories **)
(*   Inductive TypeC: Set := *)
(*     | TypeConstr     : ConstrKind -> TypeC *)
(*     | Type_          : TypeC *)
(*     | TypeTuple      : TupleKind -> TypeC *)
(*     | TypeParam      : TypeC *)
(*     | TypeSection    : TypeC *)
(*     | TypeOption     : TypeC *)
(*     | TypeParamList  : TypeC *)
(*     | TypeConstrList : ConstrKindList -> TypeC *)
(* . *)

  (** Type categories **)
  Inductive TypeC: Set :=
    | TypeConstr     : KindS -> TypeC
    | Type_          : TypeC
    | TypeTuple      : KindS -> TypeC
    | TypeParam      : TypeC
    | TypeSection    : TypeC
    | TypeOption     : TypeC
    | TypeParamList  : TypeC
    | TypeConstrList : KindS -> TypeC
  .

  (** Builder: create a class type C<Ks> **)
  Definition Class_type (C: ClassSym) (R: TypeS): TypeS :=
    (Type_apply (Type_class C) R)
  .

  (** Builder: create a tuple type X@i<R> **)
  Definition Var_type (X: TypeVar) (i:  nat) (R: TypeS): TypeS :=
    (Type_apply (Type_proj (Type_var X) i) R)
  .

  (** Builder: create a lambda-redex (<Pi> => T)<R> **)
  Definition Fun_type (Pi: TypeS) (T: TypeS) (R: TypeS): TypeS :=
    (Type_apply (Type_fun Pi T) R)
  .

  (** Property: [T] is a class type -- N **)
  Definition is_classType (T: TypeS): Prop :=
    exists C: ClassSym, exists R: TypeS,
      T = (Class_type C R)
(*     match T with *)
(*       | (Type_apply (Type_class _) _) => True *)
(*       | _                             => False *)
(*     end *)
  .

  (** Terms -- t **)
  Inductive Term: Set :=
    | Term_var    : Var -> Term                   (* method parameter, current instance *)
    | Term_select : Term -> FieldSym -> Term                         (* field selection *)
    | Term_call   : Term -> MethodSym -> TypeS -> Term -> Term       (* method call *)
    | Term_new    : TypeS -> Term -> Term                 (* instance creation *)

  (** Lists of terms -- ts **)
    | Term_nil  : Term
    | Term_cons : Term -> Term -> Term

  (** Lists of bindings from field symbols to terms -- fts **)
    | FieldTerm_nil  : Term
    | FieldTerm_cons : FieldSym -> Term -> Term -> Term
  .

  (** Class attributes **)
  Parameter classTypeParams  : ClassSym -> TypeS.
  Parameter classSuper       : ClassSym -> (option TypeS).

  (** Method attributes **)
  Parameter methodOwner      : MethodSym -> ClassSym.
  Parameter methodTypeParams : MethodSym -> TypeS.
  Parameter methodParams     : MethodSym -> (list TypeS).
  Parameter methodType       : MethodSym -> TypeS.

  (** Field attributes **)
  Parameter fieldOwner      : FieldSym -> ClassSym.
  Parameter fieldType       : FieldSym -> TypeS.

  (** Method implementations **)
  Parameter methodImpl      : MethodSym -> ClassSym -> (option Term).

  (** Main expression **)
  Parameter mainTerm        : Term.

  (** Typing context for types and terms -- D(elta) and G(amma) **)
  Definition TypeEnv: Set := (list TypeS).     (* list of Pi *)
  Definition TermEnv: Set := (list TypeS).     (* list of T *)

End Program.

