(* Author: Vincent Cremet *)
(* Date: June 2007 *) 
(* File: TermSubstitutionInTyping.v *)
(* Description: Term substitution in typing *)

(** * Term substitution in typing *)

Require Import List.
Require Import Arith.
(* Require Import Relations. *)

Require Import MatchCompNat.
Require Import Util.

Require Syntax.
Module SetProgram(My_Program: Syntax.Program).
Import My_Program.

(* Import definitions of TypeSubstitutionInTyping.v *)
Require TypeSubstitutionInTyping.
Module My_TypeSubstitutionInTyping := TypeSubstitutionInTyping.SetProgram(My_Program).
Import My_TypeSubstitutionInTyping.

Import My_BoundCompliantSubstitutions_proofs.
Import My_Progress_proofs.
Import My_TransitivityElimination_proofs.
Import My_Normalization_proofs.
Import My_Subtyping_proofs.
Import My_Typing_proofs.
Import My_Confluence.
Import My_Auxiliary_proofs.
Import My_Substitutions_proofs.
Import My_Binders_proofs.
Import My_Semantics.
Import My_WellFormedness.
Import My_Typing.
Import My_Auxiliary.
Import My_Substitutions.
Import My_Binders.

(********************************************************************)
(********************************************************************)

Inductive SubInfo (D: TypeEnv): TermInfo -> TermInfo -> Prop :=

  | SubInfo_term:
    forall (T U: TypeS),
      (Subtyping D T U) ->
      (SubInfo D (TermI T) (TermI U))

  | SubInfo_termList:
    forall (Ts Us: (list TypeS)),
      (Pointwise_subtyping D Ts Us) ->
      (SubInfo D (TermListI Ts) (TermListI Us))

  | SubInfo_fieldTermList:
    forall (N: TypeS),
      (SubInfo D (FieldTermListI N) (FieldTermListI N))
.

(********************************************************************)

Lemma SubInfo_refl:
  forall (D: TypeEnv) (info: TermInfo),
    (SubInfo D info info)
.

Proof.
intros. induction info.
(* *)
apply SubInfo_term. apply S_Refl.
(* *)
rename l into Ts. apply SubInfo_termList.
unfold Pointwise_subtyping. induction Ts.
apply Forall2_nil.
apply Forall2_cons. apply S_Refl. trivial.
(* *)
rename t into N. apply SubInfo_fieldTermList.
Qed.

Lemma SubInfo_trans:
  forall (D: TypeEnv) (info1 info2 info3: TermInfo),
    (Info_WK D info2) ->
    (SubInfo D info1 info2) ->
    (SubInfo D info2 info3) ->
    (SubInfo D info1 info3)
.

Proof.
intros. generalize info3 H1. clear info3 H1. induction H0.
(* *)
intros. inversion H1. clear info3 H1 T0 H2 H4.
simpl in H.
apply SubInfo_term. eapply S_Trans; eauto.
(* *)
intros. inversion H1. clear info3 Ts0 H1 H2 H4.  rename Us0 into Us'.
apply SubInfo_termList.
unfold Pointwise_subtyping.
eapply Forall2_trans with
  (P1 := fun T U => (Subtyping D T U) /\ (Type_WK D U Type_))
  (P2 := fun T U => (Subtyping D T U))
  (ys := Us).
intros. destruct H1. eapply S_Trans; eauto.
apply Forall2_conj with
  (P1 := fun T U => (Subtyping D T U))
  (P2 := fun (T: TypeS) U => (Type_WK D U Type_)).
apply H0.
apply List_forall_implies_Forall2_right with
  (P := fun T => (Type_WK D T Type_)).
apply H.
eapply Forall2_implies_same_length. apply H0.
apply H3.
(* *)
intros. inversion H1. apply SubInfo_fieldTermList.
Qed.

Lemma SubInfo_trans_cor1:
  forall (D: TypeEnv) (Ts1 Ts2 Ts3: (list TypeS)),
    (Info_WK D (TermListI Ts2)) ->
    (Pointwise_subtyping D Ts1 Ts2) ->
    (Pointwise_subtyping D Ts2 Ts3) ->
    (Pointwise_subtyping D Ts1 Ts3)
.

Proof.
intros.
assert (SubInfo D (TermListI Ts1) (TermListI Ts3)).
eapply SubInfo_trans. eauto. apply SubInfo_termList; trivial.
apply SubInfo_termList; trivial. inversion H2. trivial.
Qed.

(********************************************************************)

Lemma Term_substitution_preserves_fields_of_fieldTermList:
  forall (D: TypeEnv) (G: TermEnv) (N: TypeS) (fts: Term) (k: nat) (v: Term),
    (Term_WT D G fts (FieldTermListI N)) ->
    (fields_of_fieldTermList (termSubst_in_term fts k v)) =
    (fields_of_fieldTermList fts)
.

Proof.
intros. generalize H. clear H. induction fts.
intros. inversion H.
intros. inversion H.
intros. inversion H.
intros. inversion H.
intros. inversion H.
intros. inversion H.
intros. simpl. trivial.
intros. inversion H.
simpl. rewrite IHfts2. trivial. trivial.
Qed.

Lemma Term_lifting_preserves_fields_of_fieldTermList:
  forall (D: TypeEnv) (G: TermEnv) (N: TypeS) (fts: Term) (k n: nat),
    (Term_WT D G fts (FieldTermListI N)) ->
    (fields_of_fieldTermList (lift_at_depth_in_term fts k n)) =
    (fields_of_fieldTermList fts)
.

Proof.
intros. generalize H. clear H. induction fts.
intros. inversion H.
intros. inversion H.
intros. inversion H.
intros. inversion H.
intros. inversion H.
intros. inversion H.
intros. simpl. trivial.
intros. inversion H.
simpl. rewrite IHfts2. trivial. trivial.
Qed.

(********************************************************************)

Lemma Typing_right_weakening:
  forall (D: TypeEnv) (G1 G2: TermEnv) (t: Term) (info: TermInfo),
    (Term_WT D G1         t info) ->
    (Term_WT D (G1 ++ G2) (lift_in_term t (length G2)) info)
.

Proof.
intros. unfold lift_in_term. induction H.
(* T-var *)
simpl.
unfold lift_at_depth_in_var. rewrite match_comp_nat2_ge.
apply T_Var. rewrite plus_comm. apply List_from_lem2; trivial. omega.
(* T-select *)
simpl. unfold T. eapply T_Select; eauto.
(* T-call *)
simpl. unfold T. eapply T_Call; eauto.
(* T-new *)
simpl. unfold N. eapply T_New; eauto.
rewrite Term_lifting_preserves_fields_of_fieldTermList with
  (D := D) (G := G1) (N := N). trivial. trivial.
(* T-term-nil *)
simpl. apply T_term_Nil.
(* T-term-cons *)
simpl. apply T_term_Cons; trivial.
(* T-fieldTerm-nil *)
simpl. apply T_fieldTerm_Nil.
(* T-fieldTerm-cons *)
simpl. eapply T_fieldTerm_Cons; eauto.
Qed.

(********************************************************************)

Theorem Term_substitution_preserves_typing:
  forall (D: TypeEnv) (G1 G2: TypeEnv) (U U': TypeS) (u t: Term) (info: TermInfo),
    (TypeEnv_WK D) ->
    (TermEnv_WK D (G1 ++ (U :: G2))) ->
    (Term_WT D G1 u (TermI U')) ->
    (Subtyping D U' U) ->
    (Term_WT D (G1 ++ (U :: G2)) t info) ->
    exists info': TermInfo,
      (Term_WT D (G1 ++ G2) (termSubst_in_term t (length G2) u) info') /\
      (SubInfo D info' info)
.

Proof.

cut (
  forall (D: TypeEnv) (G1 G: TypeEnv) (U U': TypeS) (u t: Term) (info: TermInfo),
    (TypeEnv_WK D) ->
    (Term_WT D G1 u (TermI U')) ->
    (Subtyping D U' U) ->
    (Term_WT D G t info) ->
    forall (G2: TermEnv),
      (TermEnv_WK D (G1 ++ (U :: G2))) ->
      G = (G1 ++ (U :: G2)) ->
      exists info': TermInfo,
        (Term_WT D (G1 ++ G2) (termSubst_in_term t (length G2) u) info') /\
        (SubInfo D info' info)
).
intros. eapply H; eauto.

intros. generalize G2 H3 H4. clear G2 H3 H4.
induction H2.

(* T-var *)
intros. simpl.
(* *)
set (G' := G1 ++ U :: G2). assert (G = G'); trivial.
rewrite H5 in H2. clear G H4 H5. rename G' into G.
(* *)
unfold termSubst_in_var.
destruct (lt_eq_lt_dec i (length G2)). destruct s.
(* i < |G2| *)
rewrite match_comp_nat3_lt.
exists (TermI T). split.
(* *)
apply T_Var.
apply List_from_lem3.
apply List_from_lem4 with (xs1 := (snoc G1 U)).
replace (snoc G1 U ++ G2) with G. trivial.
unfold G. unfold snoc. rewrite app_ass. trivial.
trivial.
(* *)
apply SubInfo_term. apply S_Refl.
(* *)
trivial.
(* i = |G2| *)
rewrite match_comp_nat3_eq.
exists (TermI U').
rewrite e in H2. clear i e. unfold G in H2.
assert (T = U).
eapply List_from_is_a_partial_function. apply H2. apply List_from_lem6.
rewrite H4. split.
(* *)
apply Typing_right_weakening; trivial.
(* *)
apply SubInfo_term. trivial.
(* *)
trivial.
(* i > |G2| *)
rewrite match_comp_nat3_gt.
exists (TermI T). split.
(* *)
apply T_Var.
eapply Declaration_of_variable_1. trivial. unfold G in H2. apply H2.
(* *)
apply SubInfo_term. apply S_Refl.
(* *)
trivial.

(* T-select *)
intros. simpl.
(* *)
set (G' := G1 ++ U :: G2). assert (G = G'); trivial.
rewrite H7 in H2. rewrite H7 in IHTerm_WT. clear G H6 H7. rename G' into G.
(* *)
destruct (IHTerm_WT _ H5). trivial. destruct H6. inversion H7. rename T0 into S'.
rewrite <- H9 in H6. clear x U0 H7 H8 H9.
(* *)
exists (TermI (typeSubst_in_type T 0 R)). split.
(* *)
unfold T. eapply T_Select. eauto.
eapply S_Trans with (S := S).
eapply Typing_returns_well_kinded_types_cor1. apply H2. apply H5.
trivial. trivial. trivial.
(* *)
apply SubInfo_term. apply S_Refl.

(* T-call *)
intros. simpl.
(* *)
set (G' := G1 ++ U :: G2). assert (G = G'); trivial.
rewrite H8 in H2_. rewrite H8 in IHTerm_WT1. rewrite H8 in IHTerm_WT2.
rewrite H8 in H2_0.
clear G H8 H9. rename G' into G.
(* *)
destruct (IHTerm_WT1 _ H7). trivial. clear IHTerm_WT1. destruct H8.
inversion H9. rename T0 into S'.
rewrite <- H11 in H8. clear x U0 H10 H11 H9.
(* *)
destruct (IHTerm_WT2 _ H7). trivial. clear IHTerm_WT2. destruct H9.
inversion H10. rename Ts0 into Ss'.
rewrite <- H13 in H9. clear x Us H10 H13 H11.
(* *)
exists (TermI (typeSubst_in_type (typeSubst_in_type T 1 R') 0 R)). split.
(* *)
unfold T. eapply T_Call. eauto. eauto. eauto.
eapply S_Trans with (S := S).
eapply Typing_returns_well_kinded_types_cor1. apply H2_. apply H7.
trivial. trivial. trivial. trivial.
(* *)
apply SubInfo_trans_cor1 with (Ts2 := Ss).
eapply Typing_returns_well_kinded_types. apply H2_0. apply H7.
trivial. trivial.
(* *)
apply SubInfo_term. apply S_Refl.

(* T-new *)
intros. simpl.
(* *)
set (G' := G1 ++ U :: G2). assert (G = G'); trivial.
rewrite H8 in H4. rewrite H8 in IHTerm_WT. clear G H7 H8. rename G' into G.
(* *)
destruct (IHTerm_WT _ H6). trivial. destruct H7. inversion H8.
rewrite <- H9 in H7. rewrite H11 in H7. clear x N0 H9 H11 H8.
exists (TermI N). split.
unfold N. apply T_New. trivial. trivial. trivial.
(* *)
rewrite (Term_substitution_preserves_fields_of_fieldTermList D G N fts (length G2) u H4).
trivial.
(* *)
apply SubInfo_term. apply S_Refl.

(* T-term-nil *)
intros. simpl.
exists (TermListI nil). split.
apply T_term_Nil.
apply SubInfo_termList. unfold Pointwise_subtyping. apply Forall2_nil.

(* T-term-cons *)
intros. simpl.
(* *)
set (G' := G1 ++ U :: G2). assert (G = G'); trivial.
rewrite H2 in H2_. rewrite H2 in IHTerm_WT1. rewrite H2 in H2_0. rewrite H2 in IHTerm_WT2.
clear G H4 H2. rename G' into G.
(* *)
destruct (IHTerm_WT1 _ H3). trivial. destruct H2. inversion H4.
rewrite <- H6 in H2. rename T0 into T'. clear x U0 H6 H5 H4. clear IHTerm_WT1.
(* *)
destruct (IHTerm_WT2 _ H3). trivial. destruct H4. inversion H5.
rewrite <- H8 in H4. rename Ts0 into Ts'. clear x Us H8 H6 H5.
exists (TermListI (T' :: Ts')). split.
(* *)
apply T_term_Cons; trivial.
(* *)
apply SubInfo_termList.
unfold Pointwise_subtyping. apply Forall2_cons; trivial.

(* T-fieldTerm-nil *)
intros. simpl.
exists (FieldTermListI N). split.
apply T_fieldTerm_Nil.
apply SubInfo_fieldTermList.

(* T-fieldTerm-cons *)
intros. simpl.
(* *)
set (G' := G1 ++ U :: G2). assert (G = G'); trivial.
rewrite H7 in H2_. rewrite H7 in IHTerm_WT1. rewrite H7 in H2_0. rewrite H7 in IHTerm_WT2.
clear G H7 H8. rename G' into G.
(* *)
destruct (IHTerm_WT1 _ H6). trivial. destruct H7. inversion H8.
rewrite <- H10 in H7. rename T0 into S'. clear x U0 H10 H9 H8. clear IHTerm_WT1.
(* *)
destruct (IHTerm_WT2 _ H6). trivial. destruct H8. inversion H9.
rewrite <- H10 in H8. rewrite H13 in H8. clear x N0 H10 H13 H9.
(* *)
exists (FieldTermListI N). split.
(* *)
eapply T_fieldTerm_Cons. eauto. eauto. eauto. eauto.
eapply S_Trans with (S := S).
eapply Typing_returns_well_kinded_types_cor1. apply H2_. apply H6.
trivial. trivial. trivial.
(* *)
apply SubInfo_fieldTermList.
Qed.

(***********************************************************************)

Lemma polySubst_by_snoc:
  forall (t u: Term) (k: nat) (us: (list Term)),
    (polyTermSubst_in_term t k (snoc us u)) =
    (termSubst_in_term (polyTermSubst_in_term t (S k) us) k u)
.

Proof.
intros. generalize t u k. clear t u k.
induction us.
(* nil *)
intros. simpl. replace (k + 0) with k. trivial. omega.
(* cons *)
intros. simpl. rewrite length_of_append. simpl.
replace (k + (length us + 1)) with (S (k + length us)).
rewrite <- IHus. trivial. omega.
Qed.

(********************************************************************)

(** Poly-substitution preserves typing **)
Theorem Poly_term_substitution_preserves_typing:
  forall (D: TypeEnv) (G1 G2: TypeEnv) (Us Us': (list TypeS)) (t: Term)
         (us: (list Term)) (info: TermInfo),
    (TypeEnv_WK D) ->
    (TermEnv_WK D (G1 ++ Us ++ G2)) ->
    (Forall2 (fun u U' => (Term_WT D G1 u (TermI U'))) us Us') ->
    (Pointwise_subtyping D Us' Us) ->
    (Term_WT D (G1 ++ Us ++ G2) t info) ->
    exists info': TermInfo,
      (Term_WT D (G1 ++ G2) (polyTermSubst_in_term t (length G2) us) info') /\
      (SubInfo D info' info)
.

Proof.
intros. generalize t Us Us' info H0 H1 H2 H3. clear t Us Us' info H0 H1 H2 H3.
induction us.
(* nil *)
intros. simpl. inversion H1. rewrite <- H5 in H2.
inversion H2. rewrite <- H6 in H3. simpl in H3.
exists info. split. trivial. apply SubInfo_refl.
(* u :: us *)
rename a into u. intros.
simpl. inversion H1. clear xA xsA H4 H5. rename Us' into Ss. rename xB into U'.
rename xsB into Us'. rewrite <- H7 in H2. clear Ss H1 H7.
inversion H2. clear xA xsA H1 H4. rename Us into Ss. rename xB into U.
rename xsB into Us.
rewrite <- H7 in H0.  rewrite <- H7 in H3. clear Ss H2 H7.
(* *)
assert (TermEnv_WK D (G1 ++ Us ++ G2)).
unfold TermEnv_WK.
eapply List_forall_of_append_cor1. simpl in H0. apply H0.
(* *)
simpl in H3.
simpl in H0.
destruct (Term_substitution_preserves_typing _ _ _ _ _ _ _ _ H H0 H6 H5 H3).
destruct H2. rename x into info'.
rewrite length_of_append in H2.
(* *)
destruct (IHus _ _ _ _ H1 H8 H9 H2). destruct H7. rename x into info''.
(* *)
exists info''. split.
(* *)
rewrite plus_comm.
replace (length us) with (length Us).
trivial.
(* *)
replace (length Us) with (length Us').
symmetry. eapply Forall2_implies_same_length; eauto.
eapply Forall2_implies_same_length; eauto.
(* *)
apply SubInfo_trans with (info2 := info').
eapply Typing_returns_well_kinded_types. apply H2. apply H1.
trivial. trivial.
Qed.

(** Poly-substitution preserves typing (corollary 1) **)
Lemma Poly_term_substitution_preserves_typing_cor1:
  forall (D: TypeEnv) (G1: TypeEnv) (Us Us': (list TypeS)) (t: Term)
         (us: (list Term)) (info: TermInfo),
    (TypeEnv_WK D) ->
    (TermEnv_WK D (G1 ++ Us)) ->
    (Forall2 (fun u U' => (Term_WT D G1 u (TermI U'))) us Us') ->
    (Pointwise_subtyping D Us' Us) ->
    (Term_WT D (G1 ++ Us) t info) ->
    exists info': TermInfo,
      (Term_WT D G1 (polyTermSubst_in_term t 0 us) info') /\
      (SubInfo D info' info)
.

Proof.
intros.
replace 0 with (length (nil (A := TypeS))).
replace G1 with (G1 ++ (nil (A := TypeS))).
eapply Poly_term_substitution_preserves_typing. trivial.
replace Us with (Us ++ nil) in H0. apply H0. rewrite <- app_nil_end. trivial.
apply H1. apply H2.
replace (Us ++ nil) with Us. trivial.
rewrite <- app_nil_end. trivial.
rewrite <- app_nil_end. trivial.
trivial.
Qed.

(** Poly-substitution preserves typing (corollary 2) **)
Lemma Poly_term_substitution_preserves_typing_cor2:
  forall (Us Us': (list TypeS)) (t: Term) (us: (list Term)) (T: TypeS),
    (TermEnv_WK nil Us) ->
    (Forall2 (fun u U' => (Term_WT nil nil u (TermI U'))) us Us') ->
    (Pointwise_subtyping nil Us' Us) ->
    (Term_WT nil Us t (TermI T)) ->
    exists T': TypeS,
      (Term_WT nil nil (polyTermSubst_in_term t 0 us) (TermI T')) /\
      (Subtyping nil T' T)
.

Proof.
intros.
(* *)
assert (Hy1: (TypeEnv_WK nil)). constructor.
(* *)
assert (Hy2: (TermEnv_WK nil (nil ++ Us ++ nil))).
rewrite <- app_nil_end. trivial.
(* *)
assert (Hy3: (Term_WT nil (nil ++ Us ++ nil) t (TermI T))).
rewrite <- app_nil_end. trivial.
(* *)
destruct (Poly_term_substitution_preserves_typing _ _ _ _ _ _ _ _ Hy1 Hy2 H0 H1 Hy3).
rename x into info'. destruct H3. inversion H4. rename T0 into T'.
rewrite <- H6 in H3. simpl in H3.
(* *)
exists T'; tauto.
Qed.

End SetProgram.
