(* Author: Vincent Cremet *)
(* Date: July 2006 *) 
(* File: WellFoundation.v *)
(* Description: Lemmas about well-founded relations *)

(** * Some Properties of Well-foundation *)

Require Import List.

(** Accessible => no direct cycles **)
Lemma well_foundation1:
  forall (A: Set) (R: A -> A -> Prop) (x: A),
    (Acc R x) ->
    forall (y: A),
      (R x y) ->
      (R y x) -> False.

Proof.
intros A R x H.
induction H.
intros.
apply (H0 y H2 x H2 H1).
Qed.

(** Well-founded => no direct cycles **)
Lemma well_foundation2:
  forall (A: Set) (R: A -> A -> Prop),
    (well_founded R) ->
    forall (x y: A),
      (R x y) ->
      (R y x) -> False.

Proof.
intros.
unfold well_founded in H.
assert (Acc R x). apply H. clear H.
apply (well_foundation1 A R x H2 y H0 H1).
Qed.

(** Function extension of a binary relation **)
Definition fun_extension (A B: Set) (f: A -> B) (R: B -> B -> Prop) (x y: A): Prop :=
  (R (f x) (f y))
.

(** accessible interpretation **)
Lemma well_foundation3:
  forall (A B: Set) (f: A -> B) (R: B -> B -> Prop),
    forall (x: B),
      (Acc R x) ->
      forall (x': A),
        (x = (f x')) ->
        (Acc (fun_extension A B f R) x')
.

Proof.
intros A B f R x H.
induction H.
intros.
apply Acc_intro.
intros.
red in H2.
rewrite -> H1 in H.
rewrite -> H1 in H0. clear H1.
assert (Acc R (f y)).
apply H. trivial.
apply (H0 (f y)). trivial.
trivial.
Qed.

(** Well-founded interpretation **)
Lemma well_foundation4:
  forall (A B: Set) (f: A -> B) (R: B -> B -> Prop),
    (well_founded R) -> 
    (well_founded (fun_extension A B f R))
.

Proof.
intros.
red in H.
red.
intro.
apply (well_foundation3 A B f R (f a)).
trivial. trivial.
Qed.

(************************************************************************)
(* List_forall *)
(************************************************************************)

Section List_forall_section.

Variable A: Set.

Inductive List_forall (P: A -> Prop): (list A) -> Prop :=

  | List_forall_nil:
    (List_forall P nil)

  | List_forall_cons:
    forall (x: A) (xs: (list A)),
      (P x) ->
      (List_forall P xs) ->
      (List_forall P (x :: xs))
.

Lemma forall_implies_forall:
  forall (P: A -> Prop),
    (forall (x: A), (P x)) ->
    forall (xs: (list A)),
      (List_forall P xs)
.

Proof.
intros. induction xs.
apply List_forall_nil.
apply List_forall_cons. apply H. trivial.
Qed.

Lemma List_forall3:
  forall (P: A -> Prop) (x y z: A),
    (P x) -> (P y) -> (P z) ->
    (List_forall P (x :: y :: z :: nil))
. 

Proof.
intros.
apply List_forall_cons. trivial.
apply List_forall_cons. trivial.
apply List_forall_cons. trivial.
apply List_forall_nil.
Qed.

End List_forall_section.

Implicit Arguments List_forall [A].

(************************************************************************)
(* Sequence extension of a well-founded relation *)
(************************************************************************)

Section SequenceExtension.

Variable A: Set.
Variable R: A -> A -> Prop.

Inductive RList: (list A) -> (list A) -> Prop :=

  | RList_head:
    forall (x y: A) (xs: (list A)),
      (R x y) ->
      (RList (x :: xs) (y :: xs))

  | RList_tail:
    forall (x: A) (xs ys: (list A)),
      (RList xs ys) ->
      (RList (x :: xs) (x :: ys))
.

Lemma sequence_extension_preserves_accessibility_aux:
  forall (x: A) (xs: (list A)),
    (Acc R x) ->
    (Acc (RList) xs) ->
    (Acc (RList) (x :: xs))
.

Proof.
cut (
  forall (x: A),
    (Acc R x) ->
    forall (xs: (list A)),
      (Acc (RList) xs) ->
      (Acc (RList) (x :: xs))
).
intros. apply H. trivial. trivial.
intros x H. induction H.
intros xs H1. induction H1.
apply Acc_intro. intros. rename y into ys.
inversion H3.
(* RList-head *)
apply H0. trivial. apply Acc_intro; trivial.
(* RList-tail *)
apply H2. trivial.
Qed.

Lemma sequence_extension_preserves_accessibility:
  forall (xs: (list A)),
    (List_forall (Acc R) xs) ->
    (Acc (RList) xs)
.

Proof.
intros. induction H.
(* nil *)
apply Acc_intro. intros. inversion H.
(* cons *)
apply sequence_extension_preserves_accessibility_aux; trivial.
Qed.

Lemma sequence_extension_preserves_well_foundedness:
  (well_founded R) ->
  (well_founded RList)
.

Proof.
unfold well_founded. intros.
apply sequence_extension_preserves_accessibility.
apply forall_implies_forall. trivial.
Qed.

Lemma RTriple1:
  forall (x x' y z: A),
    (R x x') ->
    (RList (x :: y :: z :: nil) (x' :: y :: z :: nil))
.

Proof.
intros. apply RList_head. trivial.
Qed.

Lemma RTriple2:
  forall (x y y' z: A),
    (R y y') ->
    (RList (x :: y :: z :: nil) (x :: y' :: z :: nil))
.

Proof.
intros. apply RList_tail. apply RList_head. trivial.
Qed.

Lemma RTriple3:
  forall (x y z z': A),
    (R z z') ->
    (RList (x :: y :: z :: nil) (x :: y :: z' :: nil))
.

Proof.
intros. apply RList_tail. apply RList_tail. apply RList_head. trivial.
Qed.

End SequenceExtension.

(************************************************************************)
(* Lexicographic extension of well-founded relations *)
(************************************************************************)

Section LexicographicExtension.

Variable A1: Set.
Variable R1: A1 -> A1 -> Prop.
Variable A2: Set.
Variable R2: A2 -> A2 -> Prop.

Inductive RPair: (A1 * A2) -> (A1 * A2) -> Prop :=

  | RPair_fst:
    forall (x1 y1: A1) (x2 y2: A2),
      (R1 x1 y1) ->
      (RPair (x1, x2) (y1, y2))

  | RPair_snd:
    forall (x1: A1) (x2 y2: A2),
      (R2 x2 y2) ->
      (RPair (x1, x2) (x1, y2))
.

Lemma lexicographic_extension_preserves_well_foundedness:
  (well_founded R1) ->
  (well_founded R2) ->
  (well_founded RPair)
.

Proof.
intros. unfold well_founded. intro. rename a into p.
destruct p. rename a into x1. rename a0 into x2.
assert (Acc R1 x1). apply H. generalize x2. clear x2.
induction H1. intros.
assert (Acc R2 x2). apply H0. induction H3.
apply Acc_intro. intros. inversion H5.
(* RPair-fst *)
apply H2. trivial.
(* RPair-snd *)
apply H4. trivial.
Qed.

End LexicographicExtension.
